# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2014 Wikimedia Foundation Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import logging
import os
import shutil
from unittest import mock
from zuul.lib import yamlutil as yaml

import git
import testtools

from zuul.merger.merger import MergerTree, Repo
import zuul.model
from zuul.model import MergeRequest
from tests.base import (
    BaseTestCase,
    FIXTURE_DIR,
    ZuulTestCase,
    iterate_timeout,
    okay_tracebacks,
    simple_layout,
)


class TestMergerRepo(ZuulTestCase):

    log = logging.getLogger("zuul.test.merger.repo")
    tenant_config_file = 'config/single-tenant/main.yaml'
    workspace_root = None

    def setUp(self):
        super(TestMergerRepo, self).setUp()
        self.workspace_root = os.path.join(self.test_root, 'workspace')

    def test_create_head_path(self):
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        parent_repo = git.Repo(parent_path)
        parent_repo.create_head("refs/heads/foobar")
        parent_repo.create_head("refs/heads/refs/heads/foobar")

        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')
        repo = work_repo.createRepoObject(None)
        self.assertIn('foobar', repo.branches)
        self.assertIn('refs/heads/foobar', repo.branches)
        self.assertNotIn('refs/heads/refs/heads/foobar', repo.branches)

    def test_create_head_at_char(self):
        """Test that we can create branches containing the '@' char.

        This is a regression test to make sure we are not using GitPython
        APIs that interpret the '@' as a special char.
        """
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        parent_repo = git.Repo(parent_path)
        parent_repo.create_head("refs/heads/foo@bar")

        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')
        repo = work_repo.createRepoObject(None)
        self.assertIn('foo@bar', repo.branches)

    def test_ensure_cloned(self):
        parent_path = os.path.join(self.upstream_root, 'org/project1')

        # Forge a repo having a submodule
        parent_repo = git.Repo(parent_path)
        parent_repo.git(c='protocol.file.allow=always').submodule(
            'add',
            os.path.join(self.upstream_root, 'org/project2'),
            'subdir')

        parent_repo.index.commit('Adding project2 as a submodule in subdir')
        # git 1.7.8 changed .git from being a directory to a file pointing
        # to the parent repository /.git/modules/*
        self.assertTrue(os.path.exists(
            os.path.join(parent_path, 'subdir', '.git')),
            msg='.git file in submodule should be a file')

        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')
        self.assertTrue(
            os.path.isdir(os.path.join(self.workspace_root, 'subdir')),
            msg='Cloned repository has a submodule placeholder directory')
        self.assertFalse(os.path.exists(
            os.path.join(self.workspace_root, 'subdir', '.git')),
            msg='Submodule is not initialized')

        sub_repo = Repo(
            os.path.join(self.upstream_root, 'org/project2'),
            os.path.join(self.workspace_root, 'subdir'),
            'none@example.org', 'User Name', '0', '0')
        self.assertTrue(os.path.exists(
            os.path.join(self.workspace_root, 'subdir', '.git')),
            msg='Cloned over the submodule placeholder')

        self.assertEqual(
            os.path.join(self.upstream_root, 'org/project1'),
            work_repo.createRepoObject(None).remotes[0].url,
            message="Parent clone still point to upstream project1")

        self.assertEqual(
            os.path.join(self.upstream_root, 'org/project2'),
            sub_repo.createRepoObject(None).remotes[0].url,
            message="Sub repository points to upstream project2")

    def test_repo_reset_branch_conflict(self):
        """Test correct reset with conflicting branch names"""
        parent_path = os.path.join(self.upstream_root, 'org/project1')

        parent_repo = git.Repo(parent_path)
        parent_repo.create_head("foobar")

        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')

        # Checkout branch that will be deleted from the remote repo
        work_repo.checkout("foobar")

        # Delete remote branch and create a branch that conflicts with
        # the branch checked out locally.
        parent_repo.delete_head("foobar")
        parent_repo.create_head("foobar/sub")

        work_repo.update()
        work_repo.reset()
        work_repo.checkout("foobar/sub")

        # Try the reverse conflict
        parent_path = os.path.join(self.upstream_root, 'org/project2')

        parent_repo = git.Repo(parent_path)
        parent_repo.create_head("foobar/sub")

        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')

        # Checkout branch that will be deleted from the remote repo
        work_repo.checkout("foobar/sub")

        # Delete remote branch and create a branch that conflicts with
        # the branch checked out locally.
        parent_repo.delete_head("foobar/sub")

        # Note: Before git 2.13 deleting a a ref foo/bar leaves an empty
        # directory foo behind that will block creating the reference foo
        # in the future. As a workaround we must clean up empty directories
        # in .git/refs.
        if parent_repo.git.version_info[:2] < (2, 13):
            Repo._cleanup_leaked_ref_dirs(parent_path, None, [])

        parent_repo.create_head("foobar")

        work_repo.update()
        work_repo.reset()
        work_repo.checkout("foobar")

    def test_rebase_merge_conflict_abort(self):
        """Test that a failed rebase is properly aborted and related
        directories are cleaned up."""
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        parent_repo = git.Repo(parent_path)
        parent_repo.create_head("feature")

        files = {"test.txt": "master"}
        self.create_commit("org/project1", files=files, head="master",
                           message="Add master file")

        files = {"test.txt": "feature"}
        self.create_commit("org/project1", files=files, head="feature",
                           message="Add feature file")

        work_repo = Repo(parent_path, self.workspace_root,
                         "none@example.org", "User Name", "0", "0")

        item = {"ref": "refs/heads/feature"}
        # We expect the rebase to fail because of a conflict, but the
        # rebase will be aborted.
        with testtools.ExpectedException(git.exc.GitCommandError):
            work_repo.rebaseMerge(item, "master")

        # Assert that the failed rebase doesn't leave any temporary
        # directories behind.
        self.assertFalse(
            os.path.exists(f"{work_repo.local_path}/.git/rebase-merge"))
        self.assertFalse(
            os.path.exists(f"{work_repo.local_path}/.git/rebase-apply"))

    def test_rebase_merge_conflict_reset_cleanup(self):
        """Test temporary directories of a failed rebase merge are
        removed on repo reset."""
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        parent_repo = git.Repo(parent_path)
        parent_repo.create_head("feature")

        files = {"master.txt": "master"}
        self.create_commit("org/project1", files=files, head="master",
                           message="Add master file")

        files = {"feature.txt": "feature"}
        self.create_commit("org/project1", files=files, head="feature",
                           message="Add feature file")

        work_repo = Repo(parent_path, self.workspace_root,
                         "none@example.org", "User Name", "0", "0")

        # Simulate leftovers from a failed rebase
        os.mkdir(f"{work_repo.local_path}/.git/rebase-merge")
        os.mkdir(f"{work_repo.local_path}/.git/rebase-apply")

        # Resetting the repo should clean up any leaked directories
        work_repo.reset()
        item = {"ref": "refs/heads/feature"}
        work_repo.rebaseMerge(item, "master")

    def test_squash_merge_empty(self):
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        work_repo = Repo(parent_path, self.workspace_root,
                         "none@example.org", "User Name", "0", "0")
        item = {
            "ref": "refs/heads/master",
            "number": "123",
            "patchset": "1",
        }
        work_repo.squashMerge(item)

    def test_set_refs(self):
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        remote_sha = self.create_commit('org/project1')
        self.create_branch('org/project1', 'foobar')

        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')
        repo = git.Repo(self.workspace_root)
        new_sha = repo.heads.foobar.commit.hexsha
        # Unsigned regular tags are a simple refname to a commit
        unsigned_tag = repo.create_tag('unsigned', ref=new_sha)
        unsigned_sha = unsigned_tag.object.hexsha
        # Annotated tags (and signed tags) create an annotate object that
        # has its own sha that refers to another commit sha. We must
        # handle both cases in ref packing.
        annotated_tag = repo.create_tag('annotated', ref=new_sha,
                                        message='test annotated tag')
        annotated_sha = annotated_tag.object.hexsha

        work_repo.setRefs({
            'refs/heads/master': new_sha,
            'refs/tags/unsigned': unsigned_sha,
            'refs/tags/annotated': annotated_sha,
            'refs/remotes/origin/master': new_sha,
            'refs/heads/broken': 'deadbeefdeadbeefdeadbeefdeadbeefdeadbeef',
        })
        self.assertEqual(work_repo.getBranchHead('master'), new_sha)
        self.assertIn('master', repo.remotes.origin.refs)

        # Git tags have a special packed-refs format. Check that we can
        # describe both unsigned and annotated tags successfully which implies
        # we wrote packed-refs for that tag properly
        repo.git.describe('unsigned')
        repo.git.describe('annotated')

        work_repo.setRefs({'refs/heads/master': remote_sha})
        self.assertEqual(work_repo.getBranchHead('master'), remote_sha)
        self.assertNotIn('master', repo.remotes.origin.refs)

    def test_set_remote_ref(self):
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        commit_sha = self.create_commit('org/project1')
        self.create_commit('org/project1')

        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')
        work_repo.setRemoteRef('master', commit_sha)
        # missing remote ref would be created
        work_repo.setRemoteRef('missing', commit_sha)

        repo = git.Repo(self.workspace_root)
        self.assertEqual(repo.remotes.origin.refs.master.commit.hexsha,
                         commit_sha)
        self.assertEqual(repo.remotes.origin.refs.missing.commit.hexsha,
                         commit_sha)

    @okay_tracebacks('exit code')
    def test_clone_timeout(self):
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        self.patch(git.Git, 'GIT_PYTHON_GIT_EXECUTABLE',
                   os.path.join(FIXTURE_DIR, 'fake_git.sh'))
        self.patch(Repo, 'retry_attempts', 1)
        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0',
                         git_timeout=0.001)
        # TODO: have the merger and repo classes catch fewer
        # exceptions, including this one on initialization.  For the
        # test, we try cloning again.
        with testtools.ExpectedException(git.exc.GitCommandError,
                                         r'.*exit code\(-9\)'):
            work_repo._ensure_cloned(None)

    def test_fetch_timeout(self):
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        self.patch(Repo, 'retry_attempts', 1)
        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')
        work_repo.git_timeout = 0.001
        self.patch(git.Git, 'GIT_PYTHON_GIT_EXECUTABLE',
                   os.path.join(FIXTURE_DIR, 'fake_git.sh'))
        with testtools.ExpectedException(git.exc.GitCommandError,
                                         r'.*exit code\(-9\)'):
            work_repo.update()

    @okay_tracebacks('git_fetch_error.sh')
    def test_fetch_retry(self):
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        self.patch(Repo, 'retry_interval', 1)
        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')
        self.patch(git.Git, 'GIT_PYTHON_GIT_EXECUTABLE',
                   os.path.join(FIXTURE_DIR, 'git_fetch_error.sh'))
        work_repo.update()
        # This is created on the first fetch
        self.assertTrue(os.path.exists(os.path.join(
            self.workspace_root, 'stamp1')))
        # This is created on the second fetch
        self.assertTrue(os.path.exists(os.path.join(
            self.workspace_root, 'stamp2')))

    def test_deleted_local_ref(self):
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        self.create_branch('org/project1', 'foobar')

        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')

        # Delete local ref on the cached repo. This leaves us with a remote
        # ref but no local ref anymore.
        gitrepo = git.Repo(work_repo.local_path)
        gitrepo.delete_head('foobar', force=True)

        # Delete the branch upstream.
        self.delete_branch('org/project1', 'foobar')

        # And now reset the repo again. This should not crash
        work_repo.reset()

    def test_branch_rename(self):
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        # Clone upstream so that current head is master
        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')

        # Rename master to main in upstream repo
        gitrepo = git.Repo(parent_path)
        main_branch = gitrepo.create_head('main')
        gitrepo.head.reference = main_branch
        gitrepo.delete_head(gitrepo.heads['master'], force=True)

        # And now reset the repo. This should not crash
        work_repo.reset()

    @okay_tracebacks('exit code')
    def test_broken_cache(self):
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')
        self.waitUntilSettled()

        # Break the work repo
        path = work_repo.local_path
        os.remove(os.path.join(path, '.git/HEAD'))

        # And now reset the repo again. This should not crash
        work_repo.reset()

        # Now open a cache repo and break it in a way that git.Repo is happy
        # at first but git won't be because of a broken HEAD revision.
        merger = self.executor_server.merger
        cache_repo = merger.getRepo('gerrit', 'org/project')
        with open(os.path.join(cache_repo.local_path, '.git/HEAD'), 'w'):
            pass
        cache_repo.update()

        # Now open a cache repo and break it in a way that git.Repo is happy
        # at first but git won't be because of a corrupt object file.
        #
        # To construct this we create a commit so we have a guaranteed free
        # object file, then we break it by truncating it.
        fn = os.path.join(cache_repo.local_path, 'commit_filename')
        with open(fn, 'a') as f:
            f.write("test")
        repo = cache_repo.createRepoObject(None)
        repo.index.add([fn])
        repo.index.commit('test commit')

        # Pick the first object file we find and break it
        objects_path = os.path.join(cache_repo.local_path, '.git', 'objects')
        object_dir = os.path.join(
            objects_path,
            [d for d in os.listdir(objects_path) if len(d) == 2][0])
        object_to_break = os.path.join(object_dir, os.listdir(object_dir)[0])
        self.log.error(os.stat(object_to_break))
        os.chmod(object_to_break, 644)
        with open(object_to_break, 'w'):
            pass
        os.chmod(object_to_break, 444)
        cache_repo.update()

    def test_broken_gitmodules(self):
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')
        self.waitUntilSettled()

        # Break the gitmodules with uncommited changes
        path = work_repo.local_path
        with open(os.path.join(path, '.gitmodules'), 'w') as f:
            f.write('[submodule "libfoo"]\n'
                    'path = include/foo\n'
                    '---\n'
                    'url = git://example.com/git/lib.git')

        # And now reset the repo. This should not crash
        work_repo.reset()

        # Break the gitmodules with a commit
        path = work_repo.local_path
        with open(os.path.join(path, '.gitmodules'), 'w') as f:
            f.write('[submodule "libfoo"]\n'
                    'path = include/foo\n'
                    '---\n'
                    'url = git://example.com/git/lib.git')
        git_repo = work_repo._createRepoObject(work_repo.local_path,
                                               work_repo.env)
        git_repo.git.add('.gitmodules')
        git_repo.index.commit("Broken gitmodule")
        # And now reset the repo. This should not crash
        work_repo.reset()

    def test_files_changes(self):
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        self.create_branch('org/project1', 'feature')
        files = {'feature.txt': 'feature'}
        self.create_commit('org/project1', files=files, head='feature',
                           message='Add feature file')

        # Let the master diverge from the feature branch. This new file should
        # NOT be included in the changed files list.
        files = {'must-not-be-in-changelist.txt': 'FAIL'}
        self.create_commit('org/project1', files=files, head='master',
                           message='Add master file')

        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')
        changed_files = work_repo.getFilesChanges('feature', 'master')

        self.assertEqual(sorted(['README', 'feature.txt']),
                         sorted(changed_files))

    def test_files_changes_add_and_remove_files(self):
        """
        If the changed files in previous commits are reverted in later commits,
        they should not be considered as changed in the PR.
        """
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        self.create_branch('org/project1', 'feature1')

        base_sha = git.Repo(parent_path).commit('master').hexsha

        # Let the file that is also changed in the feature branch diverge
        # in master. This change should NOT be considered in the changed
        # files list.
        files = {'to-be-deleted.txt': 'FAIL'}
        self.create_commit('org/project1', files=files, head='master',
                           message='Add master file')
        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')
        # Add a file in first commit
        files = {'to-be-deleted.txt': 'test'}
        self.create_commit('org/project1', files=files, head='feature1',
                           message='Add file')
        changed_files = work_repo.getFilesChanges('feature1', base_sha)
        self.assertEqual(sorted(['README', 'to-be-deleted.txt']),
                         sorted(changed_files))
        # Delete the file in second commit
        delete_files = ['to-be-deleted.txt']
        self.create_commit('org/project1', files={},
                           delete_files=delete_files, head='feature1',
                           message='Delete file')
        changed_files = work_repo.getFilesChanges('feature1', base_sha)
        self.assertEqual(['README'], changed_files)

    def test_files_changes_master_fork_merges(self):
        """Regression test for getFilesChanges()

        Check if correct list of changed files is listed for a messy
        branch that has a merge of a fork, with the fork including a
        merge of a new master revision.

        The previously used "git merge-base" approach did not handle this
        case correctly.
        """
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        repo = git.Repo(parent_path)

        self.create_branch('org/project1', 'messy',
                           commit_filename='messy1.txt')

        # Let time pass to reproduce the order for this error case
        commit_date = datetime.datetime.now() + datetime.timedelta(seconds=5)
        commit_date = commit_date.replace(microsecond=0).isoformat()

        # Create a commit on 'master' so we can merge it into the fork
        files = {"master.txt": "master"}
        master_ref = self.create_commit('org/project1', files=files,
                                        message="Add master.txt",
                                        commit_date=commit_date)
        repo.refs.master.commit = master_ref

        # Create a fork of the 'messy' branch and merge
        # 'master' into the fork (no fast-forward)
        repo.create_head("messy-fork")
        repo.heads["messy-fork"].commit = "messy"
        repo.head.reference = 'messy'
        repo.head.reset(index=True, working_tree=True)
        repo.git.checkout('messy-fork')
        repo.git.merge('master', no_ff=True)

        # Merge fork back into 'messy' branch (no fast-forward)
        repo.head.reference = 'messy'
        repo.head.reset(index=True, working_tree=True)
        repo.git.checkout('messy')
        repo.git.merge('messy-fork', no_ff=True)

        # Create another commit on top of 'messy'
        files = {"messy2.txt": "messy2"}
        messy_ref = self.create_commit('org/project1', files=files,
                                       head='messy', message="Add messy2.txt")
        repo.refs.messy.commit = messy_ref

        # Check that we get all changes for the 'messy' but not 'master' branch
        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')
        changed_files = work_repo.getFilesChanges('messy', 'master')
        self.assertEqual(sorted(['messy1.txt', 'messy2.txt']),
                         sorted(changed_files))

    def test_update_needed(self):
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        repo = git.Repo(parent_path)
        self.create_branch('org/project1', 'stable')

        proj_repo_state_no_update_master = {
            'refs/heads/master': repo.commit('refs/heads/master').hexsha,
        }
        proj_repo_state_no_update = {
            'refs/heads/master': repo.commit('refs/heads/master').hexsha,
            'refs/heads/stable': repo.commit('refs/heads/stable').hexsha,
        }
        repo_state_no_update = {
            'gerrit': {'org/project1': proj_repo_state_no_update}
        }

        proj_repo_state_update_ref = {
            'refs/heads/master': repo.commit('refs/heads/master').hexsha,
            'refs/heads/stable': repo.commit('refs/heads/stable').hexsha,
            # New branch based on master
            'refs/heads/test': repo.commit('refs/heads/master').hexsha,
        }
        repo_state_update_ref = {
            'gerrit': {'org/project1': proj_repo_state_update_ref}
        }

        proj_repo_state_update_rev = {
            'refs/heads/master': repo.commit('refs/heads/master').hexsha,
            # Commit changed on existing branch
            'refs/heads/stable': '1234567',
        }
        repo_state_update_rev = {
            'gerrit': {'org/project1': proj_repo_state_update_rev}
        }

        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')
        self.assertFalse(work_repo.isUpdateNeeded(
            proj_repo_state_no_update_master))
        self.assertFalse(work_repo.isUpdateNeeded(proj_repo_state_no_update))
        self.assertTrue(work_repo.isUpdateNeeded(proj_repo_state_update_ref))
        self.assertTrue(work_repo.isUpdateNeeded(proj_repo_state_update_rev))

        # Get repo and update for the first time.
        merger = self.executor_server.merger
        merger.updateRepo('gerrit', 'org/project1')
        repo = merger.getRepo('gerrit', 'org/project1')
        repo.reset()

        # Branches master and stable must exist
        self.assertEqual(['master', 'stable'], repo.getBranches())

        # Test new ref causes update
        # Now create an additional branch in the parent repo
        self.create_branch('org/project1', 'stable2')

        # Update with repo state and expect no update done
        self.log.info('Calling updateRepo with repo_state_no_update')
        merger.updateRepo('gerrit', 'org/project1',
                          repo_state=repo_state_no_update)
        repo = merger.getRepo('gerrit', 'org/project1')
        repo.reset()
        self.assertEqual(['master', 'stable'], repo.getBranches())

        # Update with repo state and expect update
        self.log.info('Calling updateRepo with repo_state_update_ref')
        merger.updateRepo('gerrit', 'org/project1',
                          repo_state=repo_state_update_ref)
        repo = merger.getRepo('gerrit', 'org/project1')
        repo.reset()
        self.assertEqual(['master', 'stable', 'stable2'], repo.getBranches())

        # Test new rev causes update
        # Now create an additional branch in the parent repo
        self.create_branch('org/project1', 'stable3')

        # Update with repo state and expect no update done
        self.log.info('Calling updateRepo with repo_state_no_update')
        merger.updateRepo('gerrit', 'org/project1',
                          repo_state=repo_state_no_update)
        repo = merger.getRepo('gerrit', 'org/project1')
        repo.reset()
        self.assertEqual(['master', 'stable', 'stable2'], repo.getBranches())

        # Update with repo state and expect update
        self.log.info('Calling updateRepo with repo_state_update_rev')
        merger.updateRepo('gerrit', 'org/project1',
                          repo_state=repo_state_update_rev)
        repo = merger.getRepo('gerrit', 'org/project1')
        repo.reset()
        self.assertEqual(['master', 'stable', 'stable2', 'stable3'],
                         repo.getBranches())

        # Make sure that we always update repos that aren't in the
        # repo_state.  Prime a second project.
        self.log.info('Calling updateRepo for project2')
        merger.updateRepo('gerrit', 'org/project2',
                          repo_state=repo_state_no_update)
        repo = merger.getRepo('gerrit', 'org/project2')
        repo.reset()
        self.assertEqual(['master'],
                         repo.getBranches())

        # Then update it, passing in a repo_state where project2 is
        # not present and ensure that we perform the update.
        self.log.info('Creating stable branch for project2')
        self.create_branch('org/project2', 'stable')
        merger.updateRepo('gerrit', 'org/project2',
                          repo_state=repo_state_no_update)
        repo = merger.getRepo('gerrit', 'org/project2')
        repo.reset()
        self.assertEqual(['master', 'stable'],
                         repo.getBranches())

    def test_garbage_collect(self):
        '''Tests that git gc doesn't prune FETCH_HEAD'''
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        repo = git.Repo(parent_path)
        change_ref = 'refs/changes/01/1'

        self.log.info('Creating a commit on %s', change_ref)
        repo.head.reference = repo.head.commit
        files = {"README": "creating fake commit\n"}
        for name, content in files.items():
            file_name = os.path.join(parent_path, name)
            with open(file_name, 'a') as f:
                f.write(content)
            repo.index.add([file_name])
        commit = repo.index.commit('Test commit')
        ref = git.refs.Reference(repo, change_ref)
        ref.set_commit(commit)

        self.log.info('Cloning parent repo')
        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')

        self.log.info('Fetch %s', change_ref)
        work_repo.fetch(change_ref)

        self.log.info('Checkout master and run garbage collection')
        work_repo_object = work_repo.createRepoObject(None)
        work_repo.checkout('master')
        result = work_repo_object.git.gc('--prune=now')
        self.log.info(result)

        self.log.info('Dereferencing FETCH_HEAD')
        commit = work_repo_object.commit('FETCH_HEAD')
        self.assertIsNotNone(commit)

    def test_delete_upstream_tag(self):
        # Test that we can delete a tag from upstream and that our
        # working dir will prune it.
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        parent_repo = git.Repo(parent_path)

        # Tag upstream
        self.addTagToRepo('org/project1', 'testtag', 'HEAD')
        commit = parent_repo.commit('testtag')

        # Update downstream and verify tag matches
        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')
        work_repo_underlying = git.Repo(work_repo.local_path)
        work_repo.update()
        result = work_repo_underlying.commit('testtag')
        self.assertEqual(commit, result)

        # Delete tag upstream
        self.delTagFromRepo('org/project1', 'testtag')

        # Update downstream and verify tag is gone
        work_repo.update()
        with testtools.ExpectedException(git.exc.BadName):
            result = work_repo_underlying.commit('testtag')

        # Make a new empty commit
        new_commit = parent_repo.index.commit('test commit')
        self.assertNotEqual(commit, new_commit)

        # Tag the new commit
        self.addTagToRepo('org/project1', 'testtag', new_commit)
        new_tag_commit = parent_repo.commit('testtag')
        self.assertEqual(new_commit, new_tag_commit)

        # Verify that the downstream tag matches
        work_repo.update()
        new_result = work_repo_underlying.commit('testtag')
        self.assertEqual(new_commit, new_result)

    def test_move_upstream_tag(self):
        # Test that if an upstream tag moves, our local copy moves
        # too.
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        parent_repo = git.Repo(parent_path)

        # Tag upstream
        self.addTagToRepo('org/project1', 'testtag', 'HEAD')
        commit = parent_repo.commit('testtag')

        # Update downstream and verify tag matches
        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')
        work_repo_underlying = git.Repo(work_repo.local_path)
        work_repo.update()
        result = work_repo_underlying.commit('testtag')
        self.assertEqual(commit, result)

        # Make an empty commit
        new_commit = parent_repo.index.commit('test commit')
        self.assertNotEqual(commit, new_commit)

        # Re-tag upstream
        self.delTagFromRepo('org/project1', 'testtag')
        self.addTagToRepo('org/project1', 'testtag', new_commit)
        new_tag_commit = parent_repo.commit('testtag')
        self.assertEqual(new_commit, new_tag_commit)

        # Verify our downstream tag has moved
        work_repo.update()
        new_result = work_repo_underlying.commit('testtag')
        self.assertEqual(new_commit, new_result)

    def test_set_remote_url_clone(self):
        """Test that we always use the new Git URL for cloning.

        This is a regression test to make sure we always use the new
        Git URL when a clone of the repo is necessary before updating
        the config.
        """
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')

        # Simulate an invalid/outdated remote URL with the repo no
        # longer existing on the file system.
        work_repo.remote_url = "file:///dev/null"
        shutil.rmtree(work_repo.local_path)

        # Setting a valid remote URL should update the attribute and
        # clone the repository.
        work_repo.setRemoteUrl(parent_path)
        self.assertEqual(work_repo.remote_url, parent_path)
        self.assertTrue(os.path.exists(work_repo.local_path))

    def test_set_remote_url_invalid(self):
        """Test that we don't store the Git URL when failing to set it.

        This is a regression test to make sure we will always update
        the Git URL after a previously failed attempt.
        """
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        work_repo = Repo(parent_path, self.workspace_root,
                         'none@example.org', 'User Name', '0', '0')

        # Set the Git remote URL to an invalid value.
        invalid_url = "file:///dev/null"
        repo = work_repo.createRepoObject(None)
        work_repo._git_set_remote_url(repo, invalid_url)
        work_repo.remote_url = invalid_url

        # Simulate a failed attempt to update the remote URL
        with mock.patch.object(work_repo, "_git_set_remote_url",
                               side_effect=RuntimeError):
            with testtools.ExpectedException(RuntimeError):
                work_repo.setRemoteUrl(parent_path)

        # Make sure we cleared out the remote URL.
        self.assertIsNone(work_repo.remote_url)

        # Setting a valid remote URL should update the attribute and
        # clone the repository.
        work_repo.setRemoteUrl(parent_path)
        self.assertEqual(work_repo.remote_url, parent_path)
        self.assertTrue(os.path.exists(work_repo.local_path))


class TestMergerWithAuthUrl(ZuulTestCase):
    config_file = 'zuul-github-driver.conf'

    git_url_with_auth = True

    @simple_layout('layouts/merging-github.yaml', driver='github')
    def test_changing_url(self):
        """
        This test checks that if getGitUrl returns different urls for the same
        repo (which happens if an access token is part of the url) then the
        remote urls are changed in the merger accordingly. This tests directly
        the merger.
        """

        merger = self.executor_server.merger
        repo = merger.getRepo('github', 'org/project')
        first_url = repo.remote_url

        repo = merger.getRepo('github', 'org/project')
        second_url = repo.remote_url

        # the urls should differ
        self.assertNotEqual(first_url, second_url)

    @simple_layout('layouts/merging-github.yaml', driver='github')
    def test_changing_url_end_to_end(self):
        """
        This test checks that if getGitUrl returns different urls for the same
        repo (which happens if an access token is part of the url) then the
        remote urls are changed in the merger accordingly. This is an end to
        end test.
        """

        A = self.fake_github.openFakePullRequest('org/project', 'master',
                                                 'PR title')
        self.fake_github.emitEvent(A.getCommentAddedEvent('merge me'))
        self.waitUntilSettled()
        self.assertTrue(A.is_merged)

        # get remote url of org/project in merger
        repo = self.executor_server.merger.repos.get('github.com/org/project')
        self.assertIsNotNone(repo)
        git_repo = git.Repo(repo.local_path)
        first_url = list(git_repo.remotes[0].urls)[0]

        B = self.fake_github.openFakePullRequest('org/project', 'master',
                                                 'PR title')
        self.fake_github.emitEvent(B.getCommentAddedEvent('merge me again'))
        self.waitUntilSettled()
        self.assertTrue(B.is_merged)

        repo = self.executor_server.merger.repos.get('github.com/org/project')
        self.assertIsNotNone(repo)
        git_repo = git.Repo(repo.local_path)
        second_url = list(git_repo.remotes[0].urls)[0]

        # the urls should differ
        self.assertNotEqual(first_url, second_url)


class TestMerger(ZuulTestCase):

    tenant_config_file = 'config/single-tenant/main.yaml'

    def getRequest(self, api, request_uuid, state=None):
        for _ in iterate_timeout(30, "cache to update"):
            req = api.getRequest(request_uuid)
            if req:
                if state is None:
                    return req
                if req.state == state:
                    return req

    @staticmethod
    def _item_from_fake_change(fake_change):
        return dict(
            number=fake_change.number,
            patchset=1,
            ref=fake_change.patchsets[0]['ref'],
            connection='gerrit',
            branch=fake_change.branch,
            project=fake_change.project,
            buildset_uuid='fake-uuid',
            merge_mode=zuul.model.MERGER_MERGE_RESOLVE,
        )

    def test_merge_multiple_items(self):
        """
        Tests that the merger merges and returns the requested file changes per
        change and in the correct order.
        """

        merger = self.executor_server.merger
        files = ['zuul.yaml', '.zuul.yaml']
        dirs = ['zuul.d', '.zuul.d']

        # Simple change A
        file_dict_a = {'zuul.d/a.yaml': 'a'}
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict_a)
        item_a = self._item_from_fake_change(A)

        # Simple change B
        file_dict_b = {'zuul.d/b.yaml': 'b'}
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B',
                                           files=file_dict_b)
        item_b = self._item_from_fake_change(B)

        # Simple change C on top of A
        file_dict_c = {'zuul.d/a.yaml': 'a-with-c'}
        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C',
                                           files=file_dict_c,
                                           parent=A.patchsets[0]['ref'])
        item_c = self._item_from_fake_change(C)

        # Change in different project
        file_dict_d = {'zuul.d/a.yaml': 'a-in-project1'}
        D = self.fake_gerrit.addFakeChange('org/project1', 'master', 'D',
                                           files=file_dict_d)
        item_d = self._item_from_fake_change(D)

        # Merge A
        result = merger.mergeChanges([item_a], files=files, dirs=dirs)
        self.assertIsNotNone(result)
        hexsha, read_files, repo_state, ret_recent, orig_commit, ops = result
        self.assertEqual(len(read_files), 1)
        self.assertEqual(read_files[0]['project'], 'org/project')
        self.assertEqual(read_files[0]['branch'], 'master')
        self.assertEqual(read_files[0]['files']['zuul.d/a.yaml'], 'a')

        # Merge A -> B
        result = merger.mergeChanges([item_a, item_b], files=files, dirs=dirs)
        self.assertIsNotNone(result)
        hexsha, read_files, repo_state, ret_recent, orig_commit, ops = result
        self.assertEqual(len(read_files), 1)
        self.assertEqual(read_files[0]['project'], 'org/project')
        self.assertEqual(read_files[0]['branch'], 'master')
        self.assertEqual(read_files[0]['files']['zuul.d/a.yaml'], 'a')
        self.assertEqual(read_files[0]['files']['zuul.d/b.yaml'], 'b')

        # Merge A -> B -> C
        result = merger.mergeChanges([item_a, item_b, item_c], files=files,
                                     dirs=dirs)
        self.assertIsNotNone(result)
        hexsha, read_files, repo_state, ret_recent, orig_commit, ops = result
        self.assertEqual(len(read_files), 1)
        self.assertEqual(read_files[0]['project'], 'org/project')
        self.assertEqual(read_files[0]['branch'], 'master')
        self.assertEqual(read_files[0]['files']['zuul.d/a.yaml'],
                         'a-with-c')
        self.assertEqual(read_files[0]['files']['zuul.d/b.yaml'], 'b')

        # Merge A -> B -> C -> D
        result = merger.mergeChanges([item_a, item_b, item_c, item_d],
                                     files=files, dirs=dirs)
        self.assertIsNotNone(result)
        hexsha, read_files, repo_state, ret_recent, orig_commit, ops = result

        self.assertEqual(len(read_files), 2)
        self.assertEqual(read_files[0]['project'], 'org/project')
        self.assertEqual(read_files[0]['branch'], 'master')
        self.assertEqual(read_files[0]['files']['zuul.d/a.yaml'],
                         'a-with-c')
        self.assertEqual(read_files[0]['files']['zuul.d/b.yaml'], 'b')
        self.assertEqual(read_files[1]['project'], 'org/project1')
        self.assertEqual(read_files[1]['branch'], 'master')
        self.assertEqual(read_files[1]['files']['zuul.d/a.yaml'],
                         'a-in-project1')

    def test_merge_temp_refs(self):
        """
        Test that the merge updates local zuul refs in order to avoid
        garbage collection of needed objects.
        """
        merger = self.executor_server.merger

        parent_path = os.path.join(self.upstream_root, 'org/project')
        parent_repo = git.Repo(parent_path)
        parent_repo.create_head("foo/bar")

        # Simple change A
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        item_a = self._item_from_fake_change(A)

        # Simple change B on branch foo/bar
        B = self.fake_gerrit.addFakeChange('org/project', 'foo/bar', 'B')
        item_b = self._item_from_fake_change(B)

        # Simple change C
        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        item_c = self._item_from_fake_change(C)

        # Merge A -> B -> C
        # TODO(corvus): remove this if we update in mergeChanges
        for item in [item_a, item_b, item_c]:
            merger.updateRepo(item['connection'], item['project'])
        result = merger.mergeChanges([item_a, item_b, item_c])
        self.assertIsNotNone(result)
        merge_state = result[3]

        cache_repo = merger.getRepo('gerrit', 'org/project')
        repo = cache_repo.createRepoObject(zuul_event_id="dummy")

        # Make sure zuul refs are updated
        foobar_zuul_ref = Repo.refNameToZuulRef("foo/bar")
        master_zuul_ref = Repo.refNameToZuulRef("master")
        ref_map = {r.path: r for r in repo.refs}
        self.assertIn(foobar_zuul_ref, ref_map)
        self.assertIn(master_zuul_ref, ref_map)

        self.assertEqual(
            ref_map[master_zuul_ref].commit.hexsha,
            merge_state[("gerrit", "org/project", "master")]
        )
        self.assertEqual(
            ref_map[foobar_zuul_ref].commit.hexsha,
            merge_state[("gerrit", "org/project", "foo/bar")]
        )

        # Delete the remote branch so a reset cleanes up the local branch
        parent_repo.delete_head('foo/bar', force=True)

        # Note: Before git 2.13 deleting a a ref foo/bar leaves an empty
        # directory foo behind that will block creating the reference foo
        # in the future. As a workaround we must clean up empty directories
        # in .git/refs.
        if parent_repo.git.version_info[:2] < (2, 13):
            Repo._cleanup_leaked_ref_dirs(parent_path, None, [])

        cache_repo.update()
        cache_repo.reset()
        self.assertNotIn(foobar_zuul_ref, [r.path for r in repo.refs])

        # Create another head 'foo' that can't be created if the 'foo/bar'
        # branch wasn't cleaned up properly
        parent_repo.create_head("foo")

        # Change B now on branch 'foo'
        B = self.fake_gerrit.addFakeChange('org/project', 'foo', 'B')
        item_b = self._item_from_fake_change(B)

        # Merge A -> B -> C
        # TODO(corvus): remove this if we update in mergeChanges
        for item in [item_a, item_b, item_c]:
            merger.updateRepo(item['connection'], item['project'])
        result = merger.mergeChanges([item_a, item_b, item_c])
        self.assertIsNotNone(result)
        merge_state = result[3]

        foo_zuul_ref = Repo.refNameToZuulRef("foo")
        ref_map = {r.path: r for r in repo.refs}

        self.assertIn(foo_zuul_ref, ref_map)
        self.assertIn(master_zuul_ref, ref_map)
        self.assertEqual(
            ref_map[master_zuul_ref].commit.hexsha,
            merge_state[("gerrit", "org/project", "master")]
        )
        self.assertEqual(
            ref_map[foo_zuul_ref].commit.hexsha,
            merge_state[("gerrit", "org/project", "foo")]
        )

    def test_stale_index_lock_cleanup(self):
        # Stop the running executor's merger. We needed it running to merge
        # things during test boostrapping but now it is just in the way.
        self.executor_server._merger_running = False
        self.executor_server.merger_loop_wake_event.set()
        self.executor_server.merger_thread.join()
        # Start a dedicated merger and do a merge to populate the repo on disk
        self._startMerger()

        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.assertEqual(A.data['status'], 'MERGED')

        # Add an index.lock file
        fpath = os.path.join(self.merger_src_root, 'review.example.com',
                             'org', 'org%2Fproject1', '.git', 'index.lock')
        with open(fpath, 'w'):
            pass
        self.assertTrue(os.path.exists(fpath))

        # This will fail if git can't modify the repo due to a stale lock file.
        B = self.fake_gerrit.addFakeChange('org/project1', 'master', 'B')
        B.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.assertEqual(B.data['status'], 'MERGED')

        self.assertFalse(os.path.exists(fpath))

    def test_update_after_ff_merge(self):
        # Test update to branch from pre existing fast forwardable commit
        # causes the branch to update
        parent_path = os.path.join(self.upstream_root, 'org/project1')
        upstream_repo = git.Repo(parent_path)

        # Get repo and update for the first time.
        merger = self.executor_server.merger
        merger.updateRepo('gerrit', 'org/project1')
        repo = merger.getRepo('gerrit', 'org/project1')

        # Branch master must exist
        self.assertEqual(['master'], repo.getBranches())
        self.log.debug("Upstream master %s",
                       upstream_repo.commit('master').hexsha)

        # Create a new change in the upstream repo
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        item_a = self._item_from_fake_change(A)
        change_sha = A.data['currentPatchSet']['revision']
        change_ref = 'refs/changes/01/1/1'

        # This will pull the upstream change into the zuul repo
        self.log.info('Merge the new change so it is present in the zuul repo')
        merger.mergeChanges([item_a], zuul_event_id='testeventid')
        repo = merger.getRepo('gerrit', 'org/project1')
        zuul_repo = git.Repo(repo.local_path)
        zuul_ref = repo.refNameToZuulRef('master')
        self.log.debug("Upstream commit %s",
                       upstream_repo.commit(change_ref).hexsha)
        self.log.debug("Zuul commit %s",
                       zuul_repo.commit(zuul_ref).hexsha)
        self.assertEqual(upstream_repo.commit(change_ref).hexsha,
                         zuul_repo.commit(zuul_ref).hexsha)
        self.assertNotEqual(upstream_repo.commit(change_ref).hexsha,
                            zuul_repo.commit('refs/heads/master').hexsha)

        # Update upstream master to point at the change commit simulating a
        # fast forward merge of a change
        upstream_repo.refs.master.commit = change_sha
        self.assertEqual(upstream_repo.commit('refs/heads/master').hexsha,
                         change_sha)
        # Construct a repo state to simulate it being created by
        # another merger.
        repo_state_update_branch_ff_rev = {
            'gerrit': {
                'org/project1': {
                    'refs/heads/master': change_sha,
                }
            }
        }
        self.log.debug("Upstream master %s",
                       upstream_repo.commit('master').hexsha)

        # This should update master
        self.log.info('Update the repo and ensure it has updated properly')
        merger.updateRepo('gerrit', 'org/project1',
                          repo_state=repo_state_update_branch_ff_rev)
        merger.checkoutBranch('gerrit', 'org/project1', 'master',
                              repo_state=repo_state_update_branch_ff_rev)
        repo = merger.getRepo('gerrit', 'org/project1')
        zuul_repo = git.Repo(repo.local_path)
        self.log.debug("Zuul master %s",
                       zuul_repo.commit('master').hexsha)

        # It's not important for the zuul ref to match; it's only used
        # to avoid garbage collection, so we don't check that here.
        self.assertEqual(upstream_repo.commit('refs/heads/master').hexsha,
                         zuul_repo.commit('refs/heads/master').hexsha)
        self.assertEqual(upstream_repo.commit(change_ref).hexsha,
                         zuul_repo.commit('refs/heads/master').hexsha)
        self.assertEqual(upstream_repo.commit(change_ref).hexsha,
                         zuul_repo.commit('HEAD').hexsha)

    def test_lost_merge_requests(self):
        # Test the cleanupLostMergeRequests method of the merger
        # client.  This is normally called from apsched from the
        # scheduler.  To exercise it, we need to produce a fake lost
        # merge request and then invoke it ourselves.

        # Stop the actual merger which will see this as garbage:
        self.executor_server._merger_running = False
        self.executor_server.merger_loop_wake_event.set()
        self.executor_server.merger_thread.join()

        merger_client = self.scheds.first.sched.merger
        merger_api = merger_client.merger_api

        # Create a fake lost merge request.  This is based on
        # test_lost_merge_requests in test_zk.

        payload = {'merge': 'test'}
        merger_api.submit(MergeRequest(
            uuid='B',
            job_type=MergeRequest.MERGE,
            build_set_uuid='BB',
            tenant_name='tenant',
            pipeline_name='check',
            event_id='1',
        ), payload)
        b = self.getRequest(merger_api, "B")

        b.state = MergeRequest.RUNNING
        merger_api.update(b)
        self.getRequest(merger_api, b.uuid, MergeRequest.RUNNING)

        # The lost_merges method should only return merges which are running
        # but not locked by any merger, in this case merge b
        lost_merge_requests = list(merger_api.lostRequests())

        self.assertEqual(1, len(lost_merge_requests))
        self.assertEqual(b.path, lost_merge_requests[0].path)

        # Exercise the cleanup code
        self.log.debug("Removing lost merge requests")
        merger_client.cleanupLostMergeRequests()

        cache = merger_api.cache._cached_objects
        for _ in iterate_timeout(30, "cache to be empty"):
            if not cache:
                break

    def test_merger_get_files_changes(self):
        self.create_branch('org/project', 'stable')
        merger = self.executor_server.merger
        merger.updateRepo('gerrit', 'org/project')
        result = merger.getFilesChanges(
            'gerrit', 'org/project', 'refs/heads/stable', 'stable')
        self.assertIsNotNone(result)


class TestMergerTree(BaseTestCase):

    def test_tree(self):
        t = MergerTree()

        t.add('/root/component')
        t.add('/root/component2')
        with testtools.ExpectedException(Exception):
            t.add('/root/component/subcomponent')
        t.add('/root/foo/bar/baz')
        with testtools.ExpectedException(Exception):
            t.add('/root/foo')


class TestMergerSchemes(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    def setUp(self):
        super().setUp()
        self.work_root = os.path.join(self.test_root, 'workspace')
        self.cache_root = os.path.join(self.test_root, 'cache')

    def _getMerger(self, work_root=None, cache_root=None, scheme=None):
        work_root = work_root or self.work_root
        return self.executor_server._getMerger(
            work_root, cache_root=cache_root, scheme=scheme)

    def _assertScheme(self, root, scheme):
        if scheme == 'unique':
            self.assertTrue(os.path.exists(
                os.path.join(root, 'review.example.com',
                             'org/org%2Fproject1')))
        else:
            self.assertFalse(os.path.exists(
                os.path.join(root, 'review.example.com',
                             'org/org%2Fproject1')))

        if scheme == 'golang':
            self.assertTrue(os.path.exists(
                os.path.join(root, 'review.example.com',
                             'org/project1')))
        else:
            self.assertFalse(os.path.exists(
                os.path.join(root, 'review.example.com',
                             'org/project1')))

        if scheme == 'flat':
            self.assertTrue(os.path.exists(
                os.path.join(root, 'project1')))
        else:
            self.assertFalse(os.path.exists(
                os.path.join(root, 'project1')))

    def test_unique_scheme(self):
        cache_merger = self._getMerger(work_root=self.cache_root)
        cache_merger.updateRepo('gerrit', 'org/project1')
        self._assertScheme(self.cache_root, 'unique')

        merger = self._getMerger(
            cache_root=self.cache_root,
            scheme=zuul.model.SCHEME_UNIQUE)
        merger.getRepo('gerrit', 'org/project1')
        self._assertScheme(self.work_root, 'unique')

    def test_golang_scheme(self):
        cache_merger = self._getMerger(work_root=self.cache_root)
        cache_merger.updateRepo('gerrit', 'org/project1')
        self._assertScheme(self.cache_root, 'unique')

        merger = self._getMerger(
            cache_root=self.cache_root,
            scheme=zuul.model.SCHEME_GOLANG)
        merger.getRepo('gerrit', 'org/project1')
        self._assertScheme(self.work_root, 'golang')

    def test_flat_scheme(self):
        cache_merger = self._getMerger(work_root=self.cache_root)
        cache_merger.updateRepo('gerrit', 'org/project1')
        self._assertScheme(self.cache_root, 'unique')

        merger = self._getMerger(
            cache_root=self.cache_root,
            scheme=zuul.model.SCHEME_FLAT)
        merger.getRepo('gerrit', 'org/project1')
        self._assertScheme(self.work_root, 'flat')

    @simple_layout('layouts/overlapping-repos.yaml')
    @okay_tracebacks('collides with')
    def test_golang_collision(self):
        merger = self._getMerger(scheme=zuul.model.SCHEME_GOLANG)
        repo = merger.getRepo('gerrit', 'component')
        self.assertIsNotNone(repo)
        repo = merger.getRepo('gerrit', 'component/subcomponent')
        self.assertIsNone(repo)

    @simple_layout('layouts/overlapping-repos.yaml')
    @okay_tracebacks('collides with')
    def test_flat_collision(self):
        merger = self._getMerger(scheme=zuul.model.SCHEME_FLAT)
        repo = merger.getRepo('gerrit', 'component')
        self.assertIsNotNone(repo)
        repo = merger.getRepo('gerrit', 'component/component')
        self.assertIsNone(repo)


class TestOverlappingRepos(ZuulTestCase):

    @simple_layout('layouts/overlapping-repos.yaml')
    def test_overlapping_repos(self):
        self.executor_server.keep_jobdir = True
        A = self.fake_gerrit.addFakeChange('component', 'master', 'A')

        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='test-job', result='SUCCESS', changes='1,1')],
            ordered=False)

        build = self.getJobFromHistory('test-job')
        jobdir_git_dir = os.path.join(build.jobdir.src_root,
                                      'component', '.git')
        self.assertTrue(os.path.exists(jobdir_git_dir))
        jobdir_git_dir = os.path.join(build.jobdir.src_root,
                                      'subcomponent', '.git')
        self.assertTrue(os.path.exists(jobdir_git_dir))

        inv_path = os.path.join(build.jobdir.root, 'ansible', 'inventory.yaml')
        with open(inv_path, 'r') as f:
            inventory = yaml.safe_load(f)
        zuul = inventory['all']['vars']['zuul']
        self.assertEqual('src/component',
                         zuul['items'][0]['project']['src_dir'])
        self.assertEqual('src/component',
                         zuul['projects']['review.example.com/component']
                         ['src_dir'])
        self.assertEqual('src/component',
                         zuul['buildset_refs'][0]['src_dir'])


class TestMergerUpgrade(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    def test_merger_upgrade(self):
        work_root = os.path.join(self.test_root, 'workspace')

        # Simulate existing repos
        org_project = os.path.join(work_root, 'review.example.com', 'org',
                                   'project', '.git')
        os.makedirs(org_project)
        scheme_file = os.path.join(work_root, '.zuul_merger_scheme')

        # Verify that an executor merger doesn't "upgrade" or write a
        # scheme file.
        self.executor_server._getMerger(
            work_root, cache_root=None, scheme=zuul.model.SCHEME_FLAT)
        self.assertTrue(os.path.exists(org_project))
        self.assertFalse(os.path.exists(scheme_file))

        # Verify that a "real" merger does upgrade.
        self.executor_server._getMerger(
            work_root, cache_root=None,
            execution_context=False)

        self.assertFalse(os.path.exists(org_project))
        self.assertTrue(os.path.exists(scheme_file))
        with open(scheme_file) as f:
            self.assertEqual(f.read().strip(), 'unique')

        # Verify that the next time it starts, we don't upgrade again.
        flag_dir = os.path.join(work_root, 'flag')
        os.makedirs(flag_dir)
        self.executor_server._getMerger(
            work_root, cache_root=None,
            execution_context=False)

        self.assertFalse(os.path.exists(org_project))
        self.assertTrue(os.path.exists(scheme_file))
        self.assertTrue(os.path.exists(flag_dir))



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import tempfile
import testtools

import zuul.cmd


FAKE_CONFIG = b'''
[DEFAULT]
foo=%(ZUUL_ENV_TEST)s
'''


class TestCmd(testtools.TestCase):
    def test_read_config_with_environment(self):
        "Test that readConfig interpolates environment vars"

        self.useFixture(fixtures.EnvironmentVariable(
            'HISTTIMEFORMAT', '%Y-%m-%dT%T%z '))
        self.useFixture(fixtures.EnvironmentVariable(
            'ZUUL_ENV_TEST', 'baz'))
        with tempfile.NamedTemporaryFile() as test_config:
            test_config.write(FAKE_CONFIG)
            test_config.flush()
            app = zuul.cmd.ZuulApp()
            app.parseArguments(['-c', test_config.name])
            app.readConfig()
            self.assertEqual('baz', app.config.get('DEFAULT', 'foo'))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tests.base import (
    ZuulTestCase,
    simple_layout,
)


class TestSupercedent(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    @simple_layout('layouts/supercedent.yaml')
    def test_supercedent(self):
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        arev = A.patchsets[-1]['revision']
        A.setMerged()
        self.fake_gerrit.addEvent(A.getRefUpdatedEvent())
        self.waitUntilSettled()

        # We should never run jobs for more than one change at a time
        self.assertEqual(len(self.builds), 1)

        # This change should be superceded by the next
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        B.setMerged()
        self.fake_gerrit.addEvent(B.getRefUpdatedEvent())
        self.waitUntilSettled()

        self.assertEqual(len(self.builds), 1)

        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        crev = C.patchsets[-1]['revision']
        C.setMerged()
        self.fake_gerrit.addEvent(C.getRefUpdatedEvent())
        self.waitUntilSettled()

        self.assertEqual(len(self.builds), 1)

        self.executor_server.hold_jobs_in_build = True
        self.orderedRelease()
        self.assertHistory([
            dict(name='post-job', result='SUCCESS', newrev=arev),
            dict(name='post-job', result='SUCCESS', newrev=crev),
        ], ordered=False)

    @simple_layout('layouts/supercedent.yaml')
    def test_supercedent_branches(self):
        self.executor_server.hold_jobs_in_build = True
        self.create_branch('org/project', 'stable')
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        arev = A.patchsets[-1]['revision']
        A.setMerged()
        self.fake_gerrit.addEvent(A.getRefUpdatedEvent())
        self.waitUntilSettled()

        self.assertEqual(len(self.builds), 1)

        # This change should not be superceded
        B = self.fake_gerrit.addFakeChange('org/project', 'stable', 'B')
        brev = B.patchsets[-1]['revision']
        B.setMerged()
        self.fake_gerrit.addEvent(B.getRefUpdatedEvent())
        self.waitUntilSettled()

        self.assertEqual(len(self.builds), 2)

        self.executor_server.hold_jobs_in_build = True
        self.orderedRelease()
        self.assertHistory([
            dict(name='post-job', result='SUCCESS', newrev=arev),
            dict(name='post-job', result='SUCCESS', newrev=brev),
        ], ordered=False)

    @simple_layout('layouts/supercedent-promote.yaml')
    def test_supercedent_promote(self):
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.setMerged()
        self.fake_gerrit.addEvent(A.getChangeMergedEvent())
        self.waitUntilSettled()

        # We should never run jobs for more than one change at a time
        self.assertEqual(len(self.builds), 1)

        # This change should be superceded by the next
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        B.setMerged()
        self.fake_gerrit.addEvent(B.getChangeMergedEvent())
        self.waitUntilSettled()

        self.assertEqual(len(self.builds), 1)

        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        C.setMerged()
        self.fake_gerrit.addEvent(C.getChangeMergedEvent())
        self.waitUntilSettled()

        self.assertEqual(len(self.builds), 1)

        self.executor_server.hold_jobs_in_build = True
        self.orderedRelease()
        self.assertHistory([
            dict(name='promote-job', result='SUCCESS', changes='1,1'),
            dict(name='promote-job', result='SUCCESS', changes='3,1'),
        ], ordered=False)


class TestSupercedentCircularDependencies(ZuulTestCase):
    config_file = "zuul-gerrit-github.conf"
    tenant_config_file = "config/circular-dependencies/main.yaml"

    @simple_layout('layouts/supercedent-circular-gerrit.yaml')
    def test_supercedent_gerrit_circular_deps(self):
        # Unlike other supercedent tests, this one operates on
        # pre-merge changes instead of post-merge refs so that we can
        # better exercise the circular dependency machinery.
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project2', 'master', 'B')

        # A <-> B (via commit-depends)
        A.data["commitMessage"] = "{}\n\nDepends-On: {}\n".format(
            A.subject, B.data["url"]
        )
        B.data["commitMessage"] = "{}\n\nDepends-On: {}\n".format(
            B.subject, A.data["url"]
        )

        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        # We should never run jobs for more than one change
        # per-project at a time
        self.assertEqual(len(self.builds), 4)

        # This pair of changes should be superceded by the next
        C = self.fake_gerrit.addFakeChange('org/project1', 'master', 'C')
        D = self.fake_gerrit.addFakeChange('org/project2', 'master', 'D')

        # C <-> D (via commit-depends)
        C.data["commitMessage"] = "{}\n\nDepends-On: {}\n".format(
            C.subject, D.data["url"]
        )
        D.data["commitMessage"] = "{}\n\nDepends-On: {}\n".format(
            D.subject, C.data["url"]
        )

        self.fake_gerrit.addEvent(C.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.fake_gerrit.addEvent(D.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        # We should never run jobs for more than one change
        # per-project at a time
        self.assertEqual(len(self.builds), 4)

        # This change should supercede C
        E = self.fake_gerrit.addFakeChange('org/project1', 'master', 'E')
        self.fake_gerrit.addEvent(E.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        # We should never run jobs for more than one change
        # per-project at a time
        self.assertEqual(len(self.builds), 4)

        self.executor_server.hold_jobs_in_build = True
        self.orderedRelease()
        self.assertHistory([
            dict(name='post-job', result='SUCCESS', changes="1,1"),
            dict(name='post1-job', result='SUCCESS', changes="1,1"),
            dict(name='post-job', result='SUCCESS', changes="2,1"),
            dict(name='post2-job', result='SUCCESS', changes="2,1"),
            dict(name='post-job', result='SUCCESS', changes="5,1"),
            dict(name='post1-job', result='SUCCESS', changes="5,1"),
            dict(name='post-job', result='SUCCESS', changes="4,1"),
            dict(name='post2-job', result='SUCCESS', changes="4,1"),
        ], ordered=False)

    @simple_layout('layouts/supercedent-circular-github.yaml', driver='github')
    def test_supercedent_github_circular_deps_merged(self):
        # We leave testing pre-merge changes to the gerrit test above.
        # In this test, we're testing post-merge change objects (not
        # refs) via github since there is a reasonable post-merge
        # pipeline configuration that triggers on those.
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_github.openFakePullRequest("org/project1", "master", "A")
        B = self.fake_github.openFakePullRequest("org/project2", "master", "B")

        # A <-> B (via PR-depends)
        A.body = "{}\n\nDepends-On: {}\n".format(
            A.subject, B.url,
        )
        B.body = "{}\n\nDepends-On: {}\n".format(
            B.subject, A.url
        )

        A.setMerged('merged')
        self.fake_github.emitEvent(A.getPullRequestClosedEvent())
        self.waitUntilSettled()
        B.setMerged('merged')
        self.fake_github.emitEvent(B.getPullRequestClosedEvent())
        self.waitUntilSettled()

        # We should never run jobs for more than one change
        # per-project at a time
        self.assertEqual(len(self.builds), 4)

        # This pair of changes should be superceded by the next
        C = self.fake_github.openFakePullRequest("org/project1", "master", "C")
        D = self.fake_github.openFakePullRequest("org/project2", "master", "D")

        # C <-> D (via PR-depends)
        C.body = "{}\n\nDepends-On: {}\n".format(
            C.subject, D.url,
        )
        D.body = "{}\n\nDepends-On: {}\n".format(
            D.subject, C.url
        )

        C.setMerged('merged')
        self.fake_github.emitEvent(C.getPullRequestClosedEvent())
        self.waitUntilSettled()
        D.setMerged('merged')
        self.fake_github.emitEvent(D.getPullRequestClosedEvent())
        self.waitUntilSettled()

        # We should never run jobs for more than one change
        # per-project at a time
        self.assertEqual(len(self.builds), 4)

        # This change should supercede C
        E = self.fake_github.openFakePullRequest("org/project1", "master", "E")
        E.setMerged('merged')
        self.fake_github.emitEvent(E.getPullRequestClosedEvent())
        self.waitUntilSettled()

        # We should never run jobs for more than one change
        # per-project at a time
        self.assertEqual(len(self.builds), 4)

        self.executor_server.hold_jobs_in_build = True
        self.orderedRelease()
        self.assertHistory([
            dict(name='post-job', result='SUCCESS',
                 changes=f"{A.number},{A.head_sha}"),
            dict(name='post1-job', result='SUCCESS',
                 changes=f"{A.number},{A.head_sha}"),
            dict(name='post-job', result='SUCCESS',
                 changes=f"{B.number},{B.head_sha}"),
            dict(name='post2-job', result='SUCCESS',
                 changes=f"{B.number},{B.head_sha}"),
            dict(name='post-job', result='SUCCESS',
                 changes=f"{E.number},{E.head_sha}"),
            dict(name='post1-job', result='SUCCESS',
                 changes=f"{E.number},{E.head_sha}"),
            dict(name='post-job', result='SUCCESS',
                 changes=f"{D.number},{D.head_sha}"),
            dict(name='post2-job', result='SUCCESS',
                 changes=f"{D.number},{D.head_sha}"),
        ], ordered=False)

    @simple_layout('layouts/supercedent-circular-github.yaml', driver='github')
    def test_supercedent_github_circular_deps_closed(self):
        # Run one change through the pipeline to force all the
        # zkobjects to be created so that the nologs check below
        # doesn't see the initial error message.
        C = self.fake_github.openFakePullRequest("org/project1", "master", "C")
        self.fake_github.emitEvent(C.getPullRequestClosedEvent())
        self.waitUntilSettled("create pipeline objects")
        # We leave testing pre-merge changes to the gerrit test above.
        # In this test, we're testing post-merge change objects (not
        # refs) via github since there is a reasonable post-merge
        # pipeline configuration that triggers on those.  This test
        # exercises the code path where the change is closed but not
        # merged (we should run no jobs, but we should also not fail
        # pipeline processing while dealing with circular deps).
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_github.openFakePullRequest("org/project1", "master", "A")
        B = self.fake_github.openFakePullRequest("org/project2", "master", "B")

        # A <-> B (via PR-depends)
        A.body = "{}\n\nDepends-On: {}\n".format(
            A.subject, B.url,
        )
        B.body = "{}\n\nDepends-On: {}\n".format(
            B.subject, A.url
        )

        with self.assertNoLogs('zuul.Scheduler-0', level='ERROR'):
            self.fake_github.emitEvent(A.getPullRequestClosedEvent())
            self.waitUntilSettled()
            self.fake_github.emitEvent(B.getPullRequestClosedEvent())
            self.waitUntilSettled()



# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os

from zuul.connection import BaseConnection


class OpenstackConnection(BaseConnection):
    driver_name = 'openstack'
    log = logging.getLogger("zuul.OpenstackConnection")

    def __init__(self, driver, connection_name, connection_config):
        super().__init__(driver, connection_name, connection_config)

        # Allow the user to specify the clouds.yaml via the config
        # setting or the environment variable.  Otherwise we will use
        # the client lib default paths.
        self.config_files = self.connection_config.get(
            'client_config_file',
            os.getenv('OS_CLIENT_CONFIG_FILE', None))

        if 'cloud' not in self.connection_config:
            raise Exception('The "cloud" parameter is required for '
                            f'OpenStack connections in {self.connection_name}')
        self.cloud_name = self.connection_config.get('cloud')

        # Rate limit: requests/second
        self.rate = self.connection_config.get('rate', 2)



# Copyright 2020 BMW Group
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import sys
import json
import logging
import threading
from collections import defaultdict

from kazoo.exceptions import NoNodeError
from kazoo.protocol.states import EventType

from zuul.zk import ZooKeeperBase
from zuul.model_api import MODEL_API


COMPONENTS_ROOT = "/zuul/components"


class GlobalRegistry:
    def __init__(self):
        self.registry = None

    def create(self, zk_client):
        if not self.registry:
            self.registry = ComponentRegistry(zk_client)
        return self.registry

    def clearRegistry(self):
        self.registry = None

    @property
    def model_api(self):
        return self.registry.model_api


COMPONENT_REGISTRY = GlobalRegistry()


class BaseComponent(ZooKeeperBase):
    """
    Read/write component object.

    This object holds an offline cache of all the component's attributes. In
    case of an failed update to ZooKeeper the local cache will still hold the
    fresh values. Updating any attribute uploads all attributes to ZooKeeper.

    This enables this object to be used as local key/value store even if the
    ZooKeeper connection got lost.

    :arg client ZooKeeperClient: The Zuul ZooKeeperClient object to use, or
         None if this should be a read-only component.
    :arg hostname str: The component's hostname (multiple components with
         the same hostname may be registered; the registry will create unique
         nodes for each).
    """

    # Component states
    INITIALIZING = "initializing"
    RUNNING = "running"
    PAUSED = "paused"
    STOPPED = "stopped"

    log = logging.getLogger("zuul.Component")
    kind = "base"

    def __init__(self, client, hostname, version=None):
        # Ensure that the content is available before setting any other
        # attribute, because our __setattr__ implementation is relying on it.
        self.__dict__["content"] = {
            "hostname": hostname,
            "state": self.STOPPED,
            "kind": self.kind,
            "version": version,
            "model_api": 0,
        }
        super().__init__(client)

        self.path = None
        self.register_lock = threading.Lock()

    def __getattr__(self, name):
        try:
            return self.content[name]
        except KeyError:
            raise AttributeError

    def __setattr__(self, name, value):
        # If the specified attribute is not part of our content dictionary,
        # fall back to the default __settattr__ behaviour.
        if name not in self.content.keys():
            return super().__setattr__(name, value)

        # Set the value in the local content dict
        self.content[name] = value

        with self.register_lock:
            if not self.path:
                self.log.error(
                    "Path is not set on this component, did you forget "
                    "to call register()?"
                )
                return

            # Update the ZooKeeper node
            content = json.dumps(self.content, sort_keys=True).encode("utf-8")
            try:
                self.kazoo_client.set(self.path, content)
            except NoNodeError:
                self.log.error("Could not update %s in ZooKeeper", self)

    def register(self, model_api=MODEL_API):
        self.content['model_api'] = model_api
        with self.register_lock:
            path = "/".join([COMPONENTS_ROOT, self.kind, self.hostname])
            self.log.info("Registering component in ZooKeeper %s", path)
            self.path = self.kazoo_client.create(
                path,
                json.dumps(self.content, sort_keys=True).encode("utf-8"),
                makepath=True,
                ephemeral=True,
                sequence=True,
            )

    def _onReconnect(self):
        self.register()

    def updateFromDict(self, data):
        self.content.update(data)

    @classmethod
    def fromDict(cls, client, hostname, data):
        component = cls(client, hostname)
        component.updateFromDict(data)
        return component

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.content}>"


class SchedulerComponent(BaseComponent):
    kind = "scheduler"


class ExecutorComponent(BaseComponent):
    kind = "executor"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.initial_state = {
            "accepting_work": False,
            "zone": None,
            "allow_unzoned": False,
            "process_merge_jobs": False,
        }
        self.content.update(self.initial_state)


class MergerComponent(BaseComponent):
    kind = "merger"


class FingerGatewayComponent(BaseComponent):
    kind = "fingergw"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.initial_state = {
            "zone": None,
            "public_port": None,
            "use_ssl": False,
        }
        self.content.update(self.initial_state)


class WebComponent(BaseComponent):
    kind = "web"


class LauncherComponent(BaseComponent):
    kind = "launcher"

    def __init__(self, *args, **kwargs):
        connection_filter = kwargs.pop('connection_filter', None)
        super().__init__(*args, **kwargs)
        self.initial_state = {
            "connection_filter": connection_filter,
        }
        self.content.update(self.initial_state)


class ComponentRegistry(ZooKeeperBase):
    """A component registry is organized like:

        /zuul/components/{kind}/{sequence}

    Where kind corresponds to one of the classes in this module, and
    sequence is a ZK sequence node with a prefix of the hostname in
    order to produce a unique id for multiple identical components
    running on the same host.  An example path:

        /zuul/components/scheduler/hostname0000000000

    Components are ephemeral nodes, and so if the actual service
    disconnects from ZK, the node will disappear.

    Component objects returned by this class are read-only; updating
    their attributes will not be reflected in ZooKeeper.
    """
    log = logging.getLogger("zuul.ComponentRegistry")

    COMPONENT_CLASSES = {
        "scheduler": SchedulerComponent,
        "executor": ExecutorComponent,
        "merger": MergerComponent,
        "fingergw": FingerGatewayComponent,
        "web": WebComponent,
        "launcher": LauncherComponent,
    }

    def __init__(self, client):
        super().__init__(client)

        self.callbacks = set()
        self.client = client
        self._component_tree = None
        # kind -> hostname -> component
        self._cached_components = defaultdict(dict)

        self.model_api = None
        # Have we initialized enough to trust the model_api
        self._init = False
        # If we are already connected when the class is instantiated, directly
        # call the onConnect callback.
        if self.client.connected:
            self._onConnect()

    def _getComponentRoot(self, kind):
        return '/'.join([COMPONENTS_ROOT, kind])

    def _getComponentPath(self, kind, hostname):
        return '/'.join([COMPONENTS_ROOT, kind, hostname])

    def _onConnect(self):
        for kind in self.COMPONENT_CLASSES.keys():
            root = self._getComponentRoot(kind)
            self.kazoo_client.ensure_path(root)
            self.kazoo_client.ChildrenWatch(
                root, self._makeComponentRootWatcher(kind))
        self._init = True
        self._updateMinimumModelApi()

    def _makeComponentRootWatcher(self, kind):
        def watch(children, event=None):
            return self._onComponentRootUpdate(kind, children, event)
        return watch

    def _onComponentRootUpdate(self, kind, children, event):
        for hostname in children:
            component = self._cached_components[kind].get(hostname)
            if not component:
                self.log.info("Noticed new %s component %s", kind, hostname)
                root = self._getComponentPath(kind, hostname)
                self.kazoo_client.DataWatch(
                    root, self._makeComponentWatcher(kind, hostname))

    def _makeComponentWatcher(self, kind, hostname):
        def watch(data, stat, event=None):
            return self._onComponentUpdate(kind, hostname, data, stat, event)
        return watch

    def _onComponentUpdate(self, kind, hostname, data, stat, event):
        if event:
            etype = event.type
        else:
            etype = None
        self.log.debug(
            "Registry %s got event %s for %s %s",
            self, etype, kind, hostname)
        if (etype in (None, EventType.CHANGED, EventType.CREATED) and
            data is not None):

            # Perform an in-place update of the cached component (if any)
            component = self._cached_components.get(kind, {}).get(hostname)
            d = json.loads(data.decode("utf-8"))

            self.log.info(
                "Component %s %s updated: %s",
                kind, hostname, d)

            if component:
                if (stat.version <= component._zstat.version):
                    # Don't update to older data
                    return
                component.updateFromDict(d)
                component._zstat = stat
            else:
                # Create a new component from the data structure
                # Get the correct kind of component
                # TODO (felix): KeyError on unknown component type?
                component_cls = self.COMPONENT_CLASSES[kind]
                # Pass in null ZK client to make a read-only component
                # instance.
                component = component_cls.fromDict(None, hostname, d)
                component.path = self._getComponentPath(kind, hostname)
                component._zstat = stat

            self._cached_components[kind][hostname] = component
            self._updateMinimumModelApi()
            self._handleCallbacks()
        elif (etype == EventType.DELETED or data is None):
            self.log.info(
                "Noticed %s component %s disappeared",
                kind, hostname)
            try:
                del self._cached_components[kind][hostname]
            except KeyError:
                # If it's already gone, don't care
                pass
            self._updateMinimumModelApi()
            self._handleCallbacks()
            # Return False to stop the datawatch
            return False

    def all(self, kind=None):
        """Returns a list of components.

        If kind is None, then a list of tuples is returned, with each tuple
        being (kind, [list of components]).

        :arg kind str: The type of component to look up in the registry, or
            None to return all kinds
        """

        if kind is None:
            return [(kind, list(components.values()))
                    for (kind, components) in self._cached_components.items()]

        # Filter the cached components for the given kind
        return self._cached_components.get(kind, {}).values()

    def getMinimumModelApi(self):
        """Get the minimum model API version of all currently connected
        components"""

        # Start with our own version in case we're the only component
        # and we haven't registered.
        version = MODEL_API
        for kind, components in self.all():
            for component in components:
                version = min(version, component.model_api)
        return version

    def _updateMinimumModelApi(self):
        if not self._init:
            return
        version = self.getMinimumModelApi()
        if version != self.model_api:
            self.log.info(f"System minimum data model version {version}; "
                          f"this component {MODEL_API}")
        if self.model_api is None:
            if version < MODEL_API:
                self.log.info("The data model version of this component is "
                              "newer than the rest of the system; this "
                              "component will operate in compatability mode "
                              "until the system is upgraded")
            elif version > MODEL_API:
                self.log.error("The data model version of this component is "
                               "older than the rest of the system; "
                               "exiting to prevent data corruption")
                sys.exit(1)
        else:
            if version > self.model_api:
                if version > MODEL_API:
                    self.log.info("The data model version of this component "
                                  "is older than other components in the "
                                  "system, so other components will operate "
                                  "in a compability mode; upgrade this "
                                  "component as soon as possible to complete "
                                  "the system upgrade")
                elif version == MODEL_API:
                    self.log.info("The rest of the system has been upgraded "
                                  "to the data model version of this "
                                  "component")
            elif version < self.model_api:
                self.log.error("A component with a data model version older "
                               "than the rest of the system has been started; "
                               "data corruption is very likely to occur.")
                # Should we exit here as well?
        self.model_api = version

    def _handleCallbacks(self):
        for cb in self.callbacks:
            try:
                cb()
            except Exception:
                self.log.exception("Error in component registry callback:")

    def registerCallback(self, func):
        self.callbacks.add(func)

    def unregisterCallback(self, func):
        self.callbacks.discard(func)



# Copyright 2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import math
import time
import uuid

from zuul import model
import zuul.driver.aws.awsendpoint
import zuul.launcher.server

from tests.base import (
    driver_config,
    simple_layout,
)
from tests.unit.test_launcher import LauncherBaseTestCase


class DummyProviderNode(model.ProviderNode, subclass_id="dummy"):
    pass


class FakeItemCache:
    def __init__(self):
        self._items = {}

    def getItem(self, item_id):
        return self._items.get(item_id)

    def getItems(self):
        return self._items.values()

    def getQuota(self, provider, include_requested=False):
        return model.QuotaInformation()

    def waitForSync(self, timeout):
        return True


class FakeLauncherApi:
    def __init__(self):
        self.nodes_cache = FakeItemCache()
        self.requests_cache = FakeItemCache()

    def stop(self):
        pass

    def getMatchingRequests(self):
        return []

    def getProviderNode(self, node_id):
        return self.nodes_cache.getItem(node_id)

    def getProviderNodes(self):
        return self.nodes_cache.getItems()

    def getMatchingProviderNodes(self):
        return []


class TestSelectProviders(LauncherBaseTestCase):

    def _makeRequest(self, labels):
        request = model.NodesetRequest()
        request._set(
            uuid=uuid.uuid4().hex,
            tenant_name='tenant-one',
            pipeline_name="check",
            buildset_uuid=uuid.uuid4().hex,
            job_uuid=uuid.uuid4().hex,
            job_name="foobar",
            labels=labels,
            priority=100,
            request_time=time.time(),
            zuul_event_id=uuid.uuid4().hex,
            span_info=None,
        )
        return request

    def _makeNode(self, label, request_id=None, state=None):
        node = DummyProviderNode()
        node._set(
            uuid=uuid.uuid4().hex,
            executor_zone=None,
            state=state,
        )
        node.assignment._set(
            request_id=request_id,
        )
        return node

    def _assertProviders(self, expected, label_providers):
        self.assertEqual(len(expected), len(label_providers))
        for i, expected_label in enumerate(expected):
            self.log.debug(
                "Label %s: %s from %s",
                i,
                label_providers[i][0].name,
                label_providers[i][1].name,
            )
        for i, expected_label in enumerate(expected):
            self.assertEqual(expected_label[0], label_providers[i][0].name)
            self.assertEqual(expected_label[1], label_providers[i][1].name)

    def setUp(self):
        super().setUp()
        test = self
        self.ordered_providers = None

        def getEndpointLimits(self):
            quotas = test.test_config.driver.test_launcher['provider_quotas']
            return model.QuotaInformation(
                default=math.inf,
                **quotas[self.name],
            )
        self.patch(zuul.driver.aws.awsprovider.AwsProvider,
                   'getEndpointLimits',
                   getEndpointLimits)

        def shuffleProviders(self, providers):
            if test.ordered_providers:
                orig_providers = {p.name: p for p in providers}
                new_providers = [
                    orig_providers[pname] for pname in test.ordered_providers]
                return new_providers
            return providers[:]
        self.patch(zuul.launcher.server.Launcher,
                   '_shuffleProviders',
                   shuffleProviders)

        self.waitUntilSettled()
        self.launcher.api.stop()
        self.launcher.api = FakeLauncherApi()

    @simple_layout('layouts/launcher-select-providers.yaml',
                   enable_nodepool=True)
    @driver_config('test_launcher', provider_quotas={
        'east': {'instances': 2},
        'west': {'instances': 1},
        'unused': {'instances': 0},
    })
    def test_select_providers(self):
        # 1 node: use first available provider
        self.ordered_providers = ['east', 'west', 'unused']
        request = self._makeRequest(['debian-normal'])
        ready_nodes = zuul.launcher.server.LabelNodeMapping()
        request_ready_nodes = self.launcher._filterUnassignedNodes(
            ready_nodes, request)
        messages = []
        label_providers = self.launcher._selectProviders(
            request, request_ready_nodes, messages)
        self.launcher._emitMessages(self.log, messages)
        self._assertProviders([
            ('debian-normal', 'east'),
        ], label_providers)

        # 2 node: use "east" because it can handle it, even though
        # it's second.
        self.ordered_providers = ['unused', 'east', 'west']
        request = self._makeRequest(['debian-normal', 'debian-normal'])
        ready_nodes = zuul.launcher.server.LabelNodeMapping()
        request_ready_nodes = self.launcher._filterUnassignedNodes(
            ready_nodes, request)
        messages = []
        label_providers = self.launcher._selectProviders(
            request, request_ready_nodes, messages)
        self.launcher._emitMessages(self.log, messages)
        self._assertProviders([
            ('debian-normal', 'east'),
            ('debian-normal', 'east'),
        ], label_providers)

        # TODO: implement the TODO regarding checking for a multi-node
        # request with a single label that exceeds capacity and then
        # perform the above test with ['east', 'west', 'unused'].
        # Since we don't check that now, the only thing we can verify
        # is whether the provider can launch any nodes of a label.

        # 2 node split: 1 from east, 1 from west
        self.ordered_providers = ['east', 'west', 'unused']
        request = self._makeRequest(['debian-small', 'debian-large'])
        ready_nodes = zuul.launcher.server.LabelNodeMapping()
        request_ready_nodes = self.launcher._filterUnassignedNodes(
            ready_nodes, request)
        messages = []
        label_providers = self.launcher._selectProviders(
            request, request_ready_nodes, messages)
        self.launcher._emitMessages(self.log, messages)
        self._assertProviders([
            ('debian-small', 'east'),
            ('debian-large', 'west'),
        ], label_providers)

        # 2 node existing fail: both nodes failover
        self.ordered_providers = ['east', 'west', 'unused']
        east = 'review.example.com%2Forg%2Fcommon-config/east'
        request = self._makeRequest(['debian-normal', 'debian-normal'])
        node1 = self._makeNode('debian-normal', request.uuid)
        request.addProviderNode(node1)
        node1._set(state=node1.State.FAILED,
                   provider=east)
        self.launcher.api.nodes_cache._items[node1.uuid] = node1
        node2 = self._makeNode('debian-normal', request.uuid)
        request.addProviderNode(node2)
        node2._set(state=node1.State.BUILDING,
                   provider=east)
        self.launcher.api.nodes_cache._items[node2.uuid] = node2
        request.updateProviderNode(0, node1, add_failed_provider=east)
        request.updateProviderNode(0, node1, add_failed_provider=east)
        ready_nodes = zuul.launcher.server.LabelNodeMapping()
        request_ready_nodes = self.launcher._filterUnassignedNodes(
            ready_nodes, request)
        messages = []
        label_providers = self.launcher._selectProviders(
            request, request_ready_nodes, messages)
        self.launcher._emitMessages(self.log, messages)
        self._assertProviders([
            ('debian-normal', 'west'),
            ('debian-normal', 'west'),
        ], label_providers)

        # 1 node existing tempfail: change provider
        self.ordered_providers = ['west', 'east', 'unused']
        east = 'review.example.com%2Forg%2Fcommon-config/east'
        request = self._makeRequest(['debian-normal'])
        node1 = self._makeNode('debian-normal', request.uuid)
        request.addProviderNode(node1)
        node1._set(state=node1.State.TEMPFAILED,
                   provider=east)
        self.launcher.api.nodes_cache._items[node1.uuid] = node1
        ready_nodes = zuul.launcher.server.LabelNodeMapping()
        request_ready_nodes = self.launcher._filterUnassignedNodes(
            ready_nodes, request)
        messages = []
        label_providers = self.launcher._selectProviders(
            request, request_ready_nodes, messages)
        self.launcher._emitMessages(self.log, messages)
        self._assertProviders([
            ('debian-normal', 'west'),
        ], label_providers)

        # 2 node existing tempfail: keep same provider
        self.ordered_providers = ['west', 'east', 'unused']
        east = 'review.example.com%2Forg%2Fcommon-config/east'
        request = self._makeRequest(['debian-normal', 'debian-normal'])
        node1 = self._makeNode('debian-normal', request.uuid)
        request.addProviderNode(node1)
        node1._set(state=node1.State.TEMPFAILED,
                   provider=east)
        self.launcher.api.nodes_cache._items[node1.uuid] = node1
        node2 = self._makeNode('debian-normal', request.uuid)
        request.addProviderNode(node2)
        node2._set(state=node1.State.BUILDING,
                   provider=east)
        self.launcher.api.nodes_cache._items[node2.uuid] = node2
        ready_nodes = zuul.launcher.server.LabelNodeMapping()
        request_ready_nodes = self.launcher._filterUnassignedNodes(
            ready_nodes, request)
        messages = []
        label_providers = self.launcher._selectProviders(
            request, request_ready_nodes, messages)
        self.launcher._emitMessages(self.log, messages)
        self._assertProviders([
            ('debian-normal', 'east'),
            ('debian-normal', 'east'),
        ], label_providers)

        # 1 node with ready node in west: use west
        self.ordered_providers = ['east', 'west', 'unused']
        west = 'review.example.com%2Forg%2Fcommon-config/west'
        request = self._makeRequest(['debian-normal'])
        node1 = self._makeNode('debian-normal')
        request.addProviderNode(node1)
        provider = self.launcher._getProvider('tenant-one', 'west')
        label = provider.labels['debian-normal']
        node1._set(
            state=node1.State.READY,
            label='debian-normal',
            connection_name='aws',
            endpoint_name='aws/aws-us-west-1',
            label_config_hash=label.config_hash,
            provider=west,
        )
        self.launcher.api.nodes_cache._items[node1.uuid] = node1
        ready_nodes = self.launcher._getUnassignedNodes()
        request_ready_nodes = self.launcher._filterUnassignedNodes(
            ready_nodes, request)
        self.assertEqual(1, len(request_ready_nodes.getNodesForLabel(
            'debian-normal')))
        messages = []
        label_providers = self.launcher._selectProviders(
            request, request_ready_nodes, messages)
        self.launcher._emitMessages(self.log, messages)
        self._assertProviders([
            ('debian-normal', 'west'),
        ], label_providers)



# Copyright 2014 Rackspace Australia
# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import configparser
import os
import re
import textwrap
import time
import types

import sqlalchemy as sa

import zuul
from zuul.driver.sql.sqlreporter import (
    DATA_LENGTH_ERROR,
    SQL_MAX_STRING_LENGTH
)
from zuul.lib import yamlutil
from tests.base import (
    AnsibleZuulTestCase,
    BaseTestCase,
    FIXTURE_DIR,
    MySQLSchemaFixture,
    PostgresqlSchemaFixture,
    ZuulTestCase,
    iterate_timeout,
    okay_tracebacks,
)


class TestConnections(ZuulTestCase):
    config_file = 'zuul-connections-same-gerrit.conf'
    tenant_config_file = 'config/zuul-connections-same-gerrit/main.yaml'

    def test_multiple_gerrit_connections(self):
        "Test multiple connections to the one gerrit"

        A = self.fake_review_gerrit.addFakeChange('org/project', 'master', 'A')
        self.addEvent('review_gerrit', A.getPatchsetCreatedEvent(1))

        self.waitUntilSettled()

        self.assertEqual(len(A.patchsets[-1]['approvals']), 1)
        self.assertEqual(A.patchsets[-1]['approvals'][0]['type'], 'Verified')
        self.assertEqual(A.patchsets[-1]['approvals'][0]['value'], '1')
        self.assertEqual(A.patchsets[-1]['approvals'][0]['by']['username'],
                         'jenkins')

        B = self.fake_review_gerrit.addFakeChange('org/project', 'master', 'B')
        self.executor_server.failJob('project-test2', B)
        self.addEvent('review_gerrit', B.getPatchsetCreatedEvent(1))

        self.waitUntilSettled()

        self.assertEqual(len(B.patchsets[-1]['approvals']), 1)
        self.assertEqual(B.patchsets[-1]['approvals'][0]['type'], 'Verified')
        self.assertEqual(B.patchsets[-1]['approvals'][0]['value'], '-1')
        self.assertEqual(B.patchsets[-1]['approvals'][0]['by']['username'],
                         'civoter')


class TestSQLConnectionMysql(ZuulTestCase):
    config_file = 'zuul-sql-driver-mysql.conf'
    tenant_config_file = 'config/sql-driver/main.yaml'
    expected_table_prefix = ''

    def _sql_tables_created(self, connection_name):
        connection = self.scheds.first.connections.connections[connection_name]
        insp = sa.inspect(connection.engine)

        table_prefix = connection.table_prefix
        self.assertEqual(self.expected_table_prefix, table_prefix)

        ref_table = table_prefix + 'zuul_ref'
        buildset_table = table_prefix + 'zuul_buildset'
        buildset_ref_table = table_prefix + 'zuul_buildset_ref'
        build_table = table_prefix + 'zuul_build'
        artifact_table = table_prefix + 'zuul_artifact'
        provides_table = table_prefix + 'zuul_provides'
        build_event_table = table_prefix + 'zuul_build_event'
        buildset_event_table = table_prefix + 'zuul_buildset_event'

        self.assertEqual(9, len(insp.get_columns(ref_table)))
        self.assertEqual(12, len(insp.get_columns(buildset_table)))
        self.assertEqual(2, len(insp.get_columns(buildset_ref_table)))
        self.assertEqual(15, len(insp.get_columns(build_table)))
        self.assertEqual(5, len(insp.get_columns(artifact_table)))
        self.assertEqual(3, len(insp.get_columns(provides_table)))
        self.assertEqual(5, len(insp.get_columns(build_event_table)))
        self.assertEqual(5, len(insp.get_columns(buildset_event_table)))

    def test_sql_tables_created(self):
        "Test the tables for storing results are created properly"
        self._sql_tables_created('database')

    def _sql_indexes_created(self, connection_name):
        connection = self.scheds.first.connections.connections[connection_name]
        insp = sa.inspect(connection.engine)

        table_prefix = connection.table_prefix
        self.assertEqual(self.expected_table_prefix, table_prefix)

        ref_table = table_prefix + 'zuul_ref'
        buildset_table = table_prefix + 'zuul_buildset'
        buildset_ref_table = table_prefix + 'zuul_buildset_ref'
        build_table = table_prefix + 'zuul_build'
        artifact_table = table_prefix + 'zuul_artifact'
        provides_table = table_prefix + 'zuul_provides'
        build_event_table = table_prefix + 'zuul_build_event'
        buildset_event_table = table_prefix + 'zuul_buildset_event'

        indexes_ref = insp.get_indexes(ref_table)
        indexes_buildset = insp.get_indexes(buildset_table)
        indexes_buildset_ref = insp.get_indexes(buildset_ref_table)
        indexes_build = insp.get_indexes(build_table)
        indexes_artifact = insp.get_indexes(artifact_table)
        indexes_provides = insp.get_indexes(provides_table)
        indexes_build_event = insp.get_indexes(build_event_table)
        indexes_buildset_event = insp.get_indexes(buildset_event_table)

        self.assertEqual(8, len(indexes_ref))
        self.assertEqual(3, len(indexes_buildset))
        self.assertEqual(2, len(indexes_buildset_ref))
        self.assertEqual(5, len(indexes_build))
        self.assertEqual(1, len(indexes_artifact))
        self.assertEqual(1, len(indexes_provides))
        self.assertEqual(1, len(indexes_build_event))
        self.assertEqual(1, len(indexes_buildset_event))

        # check if all indexes are prefixed
        if table_prefix:
            indexes = (indexes_ref + indexes_buildset + indexes_buildset_ref +
                       indexes_build + indexes_artifact + indexes_provides +
                       indexes_build_event + indexes_buildset_event)
            for index in indexes:
                self.assertTrue(index['name'].startswith(table_prefix))

    def test_sql_indexes_created(self):
        "Test the indexes are created properly"
        self._sql_indexes_created('database')

    def test_sql_results(self):
        "Test results are entered into an sql table"

        def check_results():
            # Grab the sa tables
            connection = self.scheds.first.connections.getSqlConnection()
            with connection.getSession() as db:
                buildsets = list(reversed(db.getBuildsets()))
                self.assertEqual(5, len(buildsets))

                self.assertEqual('check', buildsets[0].pipeline)
                self.assertEqual('org/project', buildsets[0].refs[0].project)
                self.assertEqual(1, buildsets[0].refs[0].change)
                self.assertEqual('1', buildsets[0].refs[0].patchset)
                self.assertEqual('SUCCESS', buildsets[0].result)
                self.assertEqual('Build succeeded.', buildsets[0].message)
                self.assertEqual('tenant-one', buildsets[0].tenant)
                self.assertEqual(
                    'https://review.example.com/%d' %
                    buildsets[0].refs[0].change,
                    buildsets[0].refs[0].ref_url)
                self.assertNotEqual(None, buildsets[0].event_id)
                self.assertNotEqual(None, buildsets[0].event_timestamp)
                self.assertNotEqual(None, buildsets[0].queue_item_uuid)

                # Check the first result, which should be the project-merge job
                self.assertEqual(
                    'project-merge', buildsets[0].builds[0].job_name)
                self.assertEqual("SUCCESS", buildsets[0].builds[0].result)
                self.assertEqual(None, buildsets[0].builds[0].log_url)
                self.assertEqual('check', buildsets[1].pipeline)
                self.assertEqual('master', buildsets[1].refs[0].branch)
                self.assertEqual('org/project', buildsets[1].refs[0].project)
                self.assertEqual(2, buildsets[1].refs[0].change)
                self.assertEqual('1', buildsets[1].refs[0].patchset)
                self.assertEqual('FAILURE', buildsets[1].result)
                self.assertEqual('Build failed.', buildsets[1].message)

                # Check the second result, which should be the project-test1
                # job which failed
                self.assertEqual(
                    'project-test1', buildsets[1].builds[1].job_name)
                self.assertEqual("FAILURE", buildsets[1].builds[1].result)
                self.assertEqual(None, buildsets[1].builds[1].log_url)

                # Check the first result, which should be the project-publish
                # job
                self.assertEqual('project-publish',
                                 buildsets[2].builds[0].job_name)
                self.assertEqual("SUCCESS", buildsets[2].builds[0].result)

                self.assertEqual(
                    'project-test1', buildsets[3].builds[1].job_name)
                self.assertEqual('NODE_FAILURE', buildsets[3].builds[1].result)
                self.assertEqual(None, buildsets[3].builds[1].log_url)
                self.assertIsNotNone(buildsets[3].builds[1].start_time)
                self.assertIsNotNone(buildsets[3].builds[1].end_time)
                self.assertGreaterEqual(
                    buildsets[3].builds[1].end_time,
                    buildsets[3].builds[1].start_time)

                # Check the paused build result
                paused_build_events = buildsets[4].builds[0].build_events

                self.assertEqual(len(paused_build_events), 2)
                pause_event = paused_build_events[0]
                resume_event = paused_build_events[1]
                self.assertEqual(
                    pause_event.event_type, "paused")
                self.assertIsNotNone(pause_event.event_time)
                self.assertIsNone(pause_event.description)
                self.assertEqual(
                    resume_event.event_type, "resumed")
                self.assertIsNotNone(resume_event.event_time)
                self.assertIsNone(resume_event.description)

                self.assertGreater(
                    resume_event.event_time, pause_event.event_time)

        self.executor_server.hold_jobs_in_build = True

        # Add a success result
        self.log.debug("Adding success FakeChange")
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.orderedRelease()
        self.waitUntilSettled()

        # Add a failed result
        self.log.debug("Adding failed FakeChange")
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')

        self.executor_server.failJob('project-test1', B)
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.orderedRelease()
        self.waitUntilSettled()

        # Add a tag result
        self.log.debug("Adding FakeTag event")
        C = self.fake_gerrit.addFakeTag('org/project', 'master', 'foo')
        self.fake_gerrit.addEvent(C)
        self.waitUntilSettled()
        self.orderedRelease()
        self.waitUntilSettled()

        # Add a node_failure result
        self.fake_nodepool.pause()
        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        C.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(C.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.orderedRelease()
        self.waitUntilSettled()
        req = self.fake_nodepool.getNodeRequests()[0]
        self.fake_nodepool.addFailRequest(req)
        self.fake_nodepool.unpause()
        self.waitUntilSettled()
        self.orderedRelease()
        self.waitUntilSettled()

        # Add a paused build result
        self.log.debug("Adding paused build result")
        D = self.fake_gerrit.addFakeChange("org/project", "master", "D")
        self.executor_server.returnData(
            "project-merge", D, {"zuul": {"pause": True}})
        self.fake_gerrit.addEvent(D.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        build = self.getBuildByName('project-merge')
        self.executor_server.release('project-merge')
        for _ in iterate_timeout(10, "Wait for project-merge to pause"):
            if build.paused:
                break
        # Build has paused we wait at least one second so that the database
        # record can increment.
        time.sleep(1)
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        check_results()

    def test_sql_results_retry_builds(self):
        "Test that retry results are entered into an sql table correctly"

        # Check the results
        def check_results():
            # Grab the sa tables
            connection = self.scheds.first.connections.getSqlConnection()
            with connection.getSession() as db:
                buildsets = db.getBuildsets()

                self.assertEqual(1, len(buildsets))
                buildset0 = buildsets[0]

                self.assertEqual('check', buildset0.pipeline)
                self.assertEqual('org/project', buildset0.refs[0].project)
                self.assertEqual(1, buildset0.refs[0].change)
                self.assertEqual('1', buildset0.refs[0].patchset)
                self.assertEqual('SUCCESS', buildset0.result)
                self.assertEqual('Build succeeded.', buildset0.message)
                self.assertEqual('tenant-one', buildset0.tenant)
                self.assertEqual(
                    'https://review.example.com/%d' % buildset0.refs[0].change,
                    buildset0.refs[0].ref_url)

                # Check the retry results
                self.assertEqual('project-merge', buildset0.builds[0].job_name)
                self.assertEqual('SUCCESS', buildset0.builds[0].result)
                self.assertTrue(buildset0.builds[0].final)

                self.assertEqual('project-test1', buildset0.builds[1].job_name)
                self.assertEqual('RETRY', buildset0.builds[1].result)
                self.assertFalse(buildset0.builds[1].final)
                self.assertEqual('project-test2', buildset0.builds[2].job_name)
                self.assertEqual('RETRY', buildset0.builds[2].result)
                self.assertFalse(buildset0.builds[2].final)

                self.assertEqual('project-test1', buildset0.builds[3].job_name)
                self.assertEqual('SUCCESS', buildset0.builds[3].result)
                self.assertTrue(buildset0.builds[3].final)
                self.assertEqual('project-test2', buildset0.builds[4].job_name)
                self.assertEqual('SUCCESS', buildset0.builds[4].result)
                self.assertTrue(buildset0.builds[4].final)

        self.executor_server.hold_jobs_in_build = True

        # Add a retry result
        self.log.debug("Adding retry FakeChange")
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        # Release the merge job (which is the dependency for the other jobs)
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        # Let both test jobs fail on the first run, so they are both run again.
        self.builds[0].requeue = True
        self.builds[1].requeue = True
        self.orderedRelease()
        self.waitUntilSettled()

        check_results()

    def test_sql_intermittent_failure(self):
        # Test that if we fail to create the buildset at the start of
        # a build, we still create it at the end.
        self.executor_server.hold_jobs_in_build = True

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        # Delete the buildset
        with self.scheds.first.connections.getSqlConnection().\
             engine.connect() as conn:

            # We don't actually need to delete the zuul_ref entry
            result = conn.execute(sa.text(
                f"delete from {self.expected_table_prefix}zuul_buildset_ref;"))
            result = conn.execute(sa.text(
                f"delete from {self.expected_table_prefix}zuul_build;"))
            result = conn.execute(sa.text(
                f"delete from {self.expected_table_prefix}"
                "zuul_buildset_event;"))
            result = conn.execute(sa.text(
                f"delete from {self.expected_table_prefix}zuul_buildset;"))
            result = conn.execute(sa.text("commit;"))

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        # Check the results
        tenant = self.scheds.first.sched.abide.tenants.get("tenant-one")
        manager = tenant.layout.pipeline_managers['check']
        reporter = self.scheds.first.connections.getSqlReporter(
            manager.pipeline)

        with self.scheds.first.connections.getSqlConnection().\
             engine.connect() as conn:

            result = conn.execute(
                sa.sql.select(reporter.connection.zuul_buildset_table)
            )

            buildsets = result.fetchall()
            self.assertEqual(1, len(buildsets))
            buildset0 = buildsets[0]

            buildset0_builds = conn.execute(
                sa.sql.select(
                    reporter.connection.zuul_build_table
                ).where(
                    reporter.connection.zuul_build_table.c.buildset_id ==
                    buildset0.id
                )
            ).fetchall()

            self.assertEqual(len(buildset0_builds), 5)

    def test_sql_retry(self):
        # Exercise the SQL retry code
        reporter = self.scheds.first.sched.sql
        reporter.test_buildset_retries = 0
        reporter.test_build_retries = 0
        reporter.retry_delay = 0

        orig_createBuildset = reporter._createBuildset
        orig_createBuild = reporter._createBuild

        def _createBuildset(*args, **kw):
            ret = orig_createBuildset(*args, **kw)
            if reporter.test_buildset_retries == 0:
                reporter.test_buildset_retries += 1
                raise sa.exc.DBAPIError(None, None, None)
            return ret

        def _createBuild(*args, **kw):
            ret = orig_createBuild(*args, **kw)
            if reporter.test_build_retries == 0:
                reporter.test_build_retries += 1
                raise sa.exc.DBAPIError(None, None, None)
            return ret

        reporter._createBuildset = _createBuildset
        reporter._createBuild = _createBuild

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        # Check the results

        self.assertEqual(reporter.test_buildset_retries, 1)
        self.assertEqual(reporter.test_build_retries, 1)

        with self.scheds.first.connections.getSqlConnection().\
             engine.connect() as conn:

            result = conn.execute(
                sa.sql.select(reporter.connection.zuul_buildset_table)
            )

            buildsets = result.fetchall()
            self.assertEqual(1, len(buildsets))
            buildset0 = buildsets[0]

            buildset0_builds = conn.execute(
                sa.sql.select(
                    reporter.connection.zuul_build_table
                ).where(
                    reporter.connection.zuul_build_table.c.buildset_id ==
                    buildset0.id
                )
            ).fetchall()

            self.assertEqual(len(buildset0_builds), 5)


class TestSQLReporterLongValues(AnsibleZuulTestCase):
    config_file = 'zuul-sql-driver-mysql.conf'
    tenant_config_file = 'config/sql-reporter-long-values/main.yaml'

    def test_sql_results_log_url_too_long(self):
        "Test builds with a long log url are stored in db"

        def check_results():
            looong_log = "http://logs.example.com/l" + ("o" * 271) + "ng"
            looong_artifact_name = "tarball_l" + ("o" * 239) + "ong_name"
            # Grab the sa tables
            connection = self.scheds.first.connections.getSqlConnection()
            with connection.getSession() as db:
                buildsets = list(db.getBuildsets())
                jobs = []
                for build in buildsets[0].builds:
                    if build.job_name == 'project-test1':
                        self.assertTrue(
                            DATA_LENGTH_ERROR % ("log URL", looong_log) in
                            build.error_detail,
                            build.error_detail)
                        self.assertEqual(
                            None,
                            build.log_url)
                        # Check if the artifact name was correctly truncated
                        self.assertEqual(
                            (looong_artifact_name[:SQL_MAX_STRING_LENGTH - 15]
                             + '... (truncated)'),
                            build.artifacts[0].name)
                        # Also check if the error was recorded with the
                        # original artifact name
                        self.assertTrue(
                            DATA_LENGTH_ERROR % ("artifact name",
                                                 looong_artifact_name) in
                            build.error_detail,
                            build.error_detail)
                        return
                    jobs.append(build.job_name)
                raise Exception(
                    'The following jobs were run: "%s"' % " ".join(jobs))

        self.executor_server.hold_jobs_in_build = True

        self.log.debug("Adding success FakeChange")
        X = self.fake_gerrit.addFakeChange('org/project', 'master', 'X')

        self.fake_gerrit.addEvent(X.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.orderedRelease()
        self.waitUntilSettled()

        self.executor_server.hold_jobs_in_build = False

        check_results()


class TestSQLConnectionPostgres(TestSQLConnectionMysql):
    config_file = 'zuul-sql-driver-postgres.conf'


class TestSQLConnectionPrefixMysql(TestSQLConnectionMysql):
    config_file = 'zuul-sql-driver-prefix-mysql.conf'
    expected_table_prefix = 'prefix_'


class TestSQLConnectionPrefixPostgres(TestSQLConnectionMysql):
    config_file = 'zuul-sql-driver-prefix-postgres.conf'
    expected_table_prefix = 'prefix_'


class TestRequiredSQLConnection(BaseTestCase):
    config = None
    connections = None

    def setUp(self):
        super().setUp()
        self.addCleanup(self.stop_connection)

    def setup_connection(self, config_file):
        self.config = configparser.ConfigParser()
        self.config.read(os.path.join(FIXTURE_DIR, config_file))

        # Setup databases
        for section_name in self.config.sections():
            con_match = re.match(r'^connection ([\'\"]?)(.*)(\1)$',
                                 section_name, re.I)
            if not con_match:
                continue

            if self.config.get(section_name, 'driver') == 'sql':
                if (self.config.get(section_name, 'dburi') ==
                        '$MYSQL_FIXTURE_DBURI$'):
                    f = MySQLSchemaFixture(self.id(), self.random_databases,
                                           self.delete_databases)
                    self.useFixture(f)
                    self.config.set(section_name, 'dburi', f.dburi)
                elif (self.config.get(section_name, 'dburi') ==
                      '$POSTGRESQL_FIXTURE_DBURI$'):
                    f = PostgresqlSchemaFixture(self.id(),
                                                self.random_databases,
                                                self.delete_databases)
                    self.useFixture(f)
                    self.config.set(section_name, 'dburi', f.dburi)
        self.connections = zuul.lib.connections.ConnectionRegistry()

    def stop_connection(self):
        self.connections.stop()


class TestMultipleGerrits(ZuulTestCase):
    config_file = 'zuul-connections-multiple-gerrits.conf'
    tenant_config_file = 'config/zuul-connections-multiple-gerrits/main.yaml'

    def test_multiple_project_separate_gerrits(self):
        self.executor_server.hold_jobs_in_build = True

        A = self.fake_another_gerrit.addFakeChange(
            'org/project1', 'master', 'A')
        self.fake_another_gerrit.addEvent(A.getPatchsetCreatedEvent(1))

        self.waitUntilSettled()

        self.assertBuilds([dict(name='project-test2',
                                changes='1,1',
                                project='org/project1',
                                pipeline='another_check')])

        # NOTE(jamielennox): the tests back the git repo for both connections
        # onto the same git repo on the file system. If we just create another
        # fake change the fake_review_gerrit will try to create another 1,1
        # change and git will fail to create the ref. Arbitrarily set it to get
        # around the problem.
        self.fake_review_gerrit.change_number = 50

        B = self.fake_review_gerrit.addFakeChange(
            'org/project1', 'master', 'B')
        self.fake_review_gerrit.addEvent(B.getPatchsetCreatedEvent(1))

        self.waitUntilSettled()

        self.assertBuilds([
            dict(name='project-test2',
                 changes='1,1',
                 project='org/project1',
                 pipeline='another_check'),
            dict(name='project-test1',
                 changes='51,1',
                 project='org/project1',
                 pipeline='review_check'),
        ])

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

    def test_multiple_project_separate_gerrits_common_pipeline(self):
        self.executor_server.hold_jobs_in_build = True

        self.create_branch('org/project2', 'develop')
        self.fake_another_gerrit.addEvent(
            self.fake_another_gerrit.getFakeBranchCreatedEvent(
                'org/project2', 'develop'))

        self.fake_another_gerrit.addEvent(
            self.fake_review_gerrit.getFakeBranchCreatedEvent(
                'org/project2', 'develop'))
        self.waitUntilSettled()

        A = self.fake_another_gerrit.addFakeChange(
            'org/project2', 'master', 'A')
        self.fake_another_gerrit.addEvent(A.getPatchsetCreatedEvent(1))

        self.waitUntilSettled()

        self.assertBuilds([dict(name='project-test2',
                                changes='1,1',
                                project='org/project2',
                                pipeline='common_check')])

        # NOTE(jamielennox): the tests back the git repo for both connections
        # onto the same git repo on the file system. If we just create another
        # fake change the fake_review_gerrit will try to create another 1,1
        # change and git will fail to create the ref. Arbitrarily set it to get
        # around the problem.
        self.fake_review_gerrit.change_number = 50

        B = self.fake_review_gerrit.addFakeChange(
            'org/project2', 'develop', 'B')
        self.fake_review_gerrit.addEvent(B.getPatchsetCreatedEvent(1))

        self.waitUntilSettled()

        self.assertBuilds([
            dict(name='project-test2',
                 changes='1,1',
                 project='org/project2',
                 pipeline='common_check'),
            dict(name='project-test1',
                 changes='51,1',
                 project='org/project2',
                 pipeline='common_check'),
        ])

        # NOTE(avass): This last change should not trigger any pipelines since
        # common_check is configured to only run on master for another_gerrit
        C = self.fake_another_gerrit.addFakeChange(
            'org/project2', 'develop', 'C')
        self.fake_another_gerrit.addEvent(C.getPatchsetCreatedEvent(1))

        self.waitUntilSettled()

        self.assertBuilds([
            dict(name='project-test2',
                 changes='1,1',
                 project='org/project2',
                 pipeline='common_check'),
            dict(name='project-test1',
                 changes='51,1',
                 project='org/project2',
                 pipeline='common_check'),
        ])

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()


class TestConnectionsMerger(ZuulTestCase):
    config_file = 'zuul-connections-merger.conf'
    tenant_config_file = 'config/single-tenant/main.yaml'

    def test_connections_merger(self):
        "Test merger only configures source connections"

        self.assertIn("gerrit", self.executor_server.connections.connections)
        self.assertIn("github", self.executor_server.connections.connections)
        self.assertNotIn("smtp", self.executor_server.connections.connections)
        self.assertNotIn("sql", self.executor_server.connections.connections)
        self.assertNotIn("timer", self.executor_server.connections.connections)
        self.assertNotIn("zuul", self.executor_server.connections.connections)


class TestConnectionsCgit(ZuulTestCase):
    config_file = 'zuul-connections-cgit.conf'
    tenant_config_file = 'config/single-tenant/main.yaml'

    def test_cgit_web_url(self):
        self.assertIn("gerrit", self.scheds.first.connections.connections)
        conn = self.scheds.first.connections.connections['gerrit']
        source = conn.source
        proj = source.getProject('foo/bar')
        url = conn._getWebUrl(proj, '1')
        self.assertEqual(url,
                         'https://cgit.example.com/cgit/foo/bar/commit/?id=1')


class TestConnectionsGitweb(ZuulTestCase):
    config_file = 'zuul-connections-gitweb.conf'
    tenant_config_file = 'config/single-tenant/main.yaml'

    def test_gitweb_url(self):
        self.assertIn("gerrit", self.scheds.first.connections.connections)
        conn = self.scheds.first.connections.connections['gerrit']
        source = conn.source
        proj = source.getProject('foo/bar')
        url = conn._getWebUrl(proj, '1')
        url_should_be = 'https://review.example.com/' \
                        'gitweb?p=foo/bar.git;a=commitdiff;h=1'
        self.assertEqual(url, url_should_be)


class TestMQTTConnection(ZuulTestCase):
    config_file = 'zuul-mqtt-driver.conf'
    tenant_config_file = 'config/mqtt-driver/main.yaml'

    @okay_tracebacks('Connection refused')
    def test_mqtt_reporter(self):
        "Test the MQTT reporter"
        # Add a success result
        A = self.fake_gerrit.addFakeChange('org/project',
                                           'master',
                                           'A',
                                           topic='fake/topic')
        artifact = {'name': 'image',
                    'url': 'http://example.com/image',
                    'metadata': {
                        'type': 'container_image'
                    }}
        self.executor_server.returnData(
            "test", A, {
                "zuul": {
                    "log_url": "some-log-url",
                    'artifacts': [artifact],
                },
                'foo': 'bar',
            }
        )
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        success_event_check = self.mqtt_messages.pop()
        start_event_check = self.mqtt_messages.pop()

        self.assertEqual(start_event_check.get('topic'),
                         'tenant-one/zuul_start/check/org/project/master')
        mqtt_payload = start_event_check['msg']
        self.assertEqual(mqtt_payload['project'], 'org/project')
        self.assertEqual(len(mqtt_payload['commit_id']), 40)
        self.assertEqual(mqtt_payload['owner'], 'username')
        self.assertEqual(mqtt_payload['branch'], 'master')
        self.assertEqual(mqtt_payload['buildset']['result'], None)
        self.assertEqual(mqtt_payload['buildset']['builds'][0]['job_name'],
                         'test')
        self.assertNotIn('result', mqtt_payload['buildset']['builds'][0])
        self.assertNotIn('artifacts', mqtt_payload['buildset']['builds'][0])
        builds = mqtt_payload['buildset']['builds']
        test_job = [b for b in builds if b['job_name'] == 'test'][0]
        self.assertNotIn('returned_data', test_job)

        self.assertEqual(success_event_check.get('topic'),
                         'tenant-one/zuul_buildset/check/org/project/master')
        mqtt_payload = success_event_check['msg']
        self.assertEqual(mqtt_payload['project'], 'org/project')
        self.assertEqual(mqtt_payload['pipeline'], 'check')
        self.assertEqual(mqtt_payload['queue'], '')
        self.assertEqual(mqtt_payload['branch'], 'master')
        self.assertEqual(mqtt_payload['buildset']['result'], 'SUCCESS')
        builds = mqtt_payload['buildset']['builds']
        test_job = [b for b in builds if b['job_name'] == 'test'][0]
        dependent_test_job = [
            b for b in builds if b['job_name'] == 'dependent-test'
        ][0]
        self.assertEqual(test_job['job_name'], 'test')
        self.assertEqual(test_job['result'], 'SUCCESS')
        self.assertEqual(test_job['dependencies'], [])
        self.assertEqual(test_job['job_dependencies'], {})
        self.assertEqual(test_job['artifacts'], [artifact])
        self.assertEqual(test_job['log_url'], 'some-log-url/')
        self.assertEqual(test_job['returned_data'], {'foo': 'bar'})
        build_id = test_job["uuid"]
        self.assertEqual(
            test_job["web_url"],
            "https://tenant.example.com/t/tenant-one/build/{}".format(
                build_id
            ),
        )
        self.assertIn('execute_time', test_job)
        self.assertIn('timestamp', mqtt_payload)
        self.assertIn('enqueue_time', mqtt_payload)
        self.assertIn('trigger_time', mqtt_payload)
        self.assertIn('zuul_event_id', mqtt_payload)
        self.assertIn('uuid', mqtt_payload)
        self.assertEqual(dependent_test_job['dependencies'], ['test'])
        self.assertIn('test', dependent_test_job['job_dependencies'])
        self.assertEqual(mqtt_payload['tenant_web_root'],
                         'https://tenant.example.com/t/tenant-one/')

        changes = mqtt_payload['changes']
        self.assertEqual(len(changes), 1)
        self.assertEqual(changes[0]['topic'], 'fake/topic')

        A.addApproval("Code-Review", 2)
        self.fake_gerrit.addEvent(A.addApproval("Approved", 1))
        self.waitUntilSettled()

        success_event_gate = self.mqtt_messages.pop()
        mqtt_payload = success_event_gate['msg']
        self.assertEqual(mqtt_payload['project'], 'org/project')
        self.assertEqual(mqtt_payload['pipeline'], 'gate')
        self.assertEqual(mqtt_payload['queue'], 'integrated')

    @okay_tracebacks('Connection refused')
    def test_mqtt_paused_job(self):
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange("org/project", "master", "A")
        # Let the job being paused via the executor
        self.executor_server.returnData("test", A, {"zuul": {"pause": True}})
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        build = self.getBuildByName('test')
        self.executor_server.release('test')
        for _ in iterate_timeout(10, "Wait for test to pause"):
            if build.paused:
                break
        # Build has paused we wait at least one second so that the timestamp
        # record can increment.
        time.sleep(1)
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        success_event = self.mqtt_messages.pop()

        mqtt_payload = success_event["msg"]
        self.assertEqual(mqtt_payload["project"], "org/project")
        builds = mqtt_payload["buildset"]["builds"]
        paused_job = [b for b in builds if b["job_name"] == "test"][0]
        self.assertEqual(len(paused_job["events"]), 2)

        pause_event = paused_job["events"][0]
        self.assertEqual(pause_event["event_type"], "paused")
        self.assertGreater(
            pause_event["event_time"], paused_job["start_time"])
        self.assertLess(pause_event["event_time"], paused_job["end_time"])

        resume_event = paused_job["events"][1]
        self.assertEqual(resume_event["event_type"], "resumed")
        self.assertGreater(
            resume_event["event_time"], paused_job["start_time"])
        self.assertLess(resume_event["event_time"], paused_job["end_time"])

        self.assertGreater(
            resume_event["event_time"], pause_event["event_time"])

    @okay_tracebacks('Connection refused')
    def test_mqtt_retried_builds(self):
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange("org/project", "master", "A")
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertEqual(len(self.builds), 1)
        build = self.builds[0]
        build.requeue = True
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        success_event = self.mqtt_messages.pop()
        mqtt_payload = success_event["msg"]
        retry_builds = mqtt_payload["buildset"]["retry_builds"]
        self.assertEqual(len(retry_builds), 1)
        retry_build = retry_builds[0]

        self.assertEqual(build.job.name, retry_build["job_name"])
        self.assertEqual(build.job.uuid, retry_build["job_uuid"])
        self.assertEqual(build.job.voting, retry_build["voting"])
        self.assertEqual(build.uuid, retry_build["uuid"])
        self.assertIn(build.uuid, retry_build["web_url"])
        self.assertEqual("RETRY", retry_build["result"])

        self.assertIn("start_time", retry_build)
        self.assertIn("end_time", retry_build)
        self.assertIn("execute_time", retry_build)
        self.assertIn("log_url", retry_build)

    @okay_tracebacks('Connection refused')
    def test_mqtt_invalid_topic(self):
        in_repo_conf = textwrap.dedent(
            """
            - pipeline:
                name: test-pipeline
                manager: independent
                trigger:
                  gerrit:
                    - event: comment-added
                start:
                  mqtt:
                    topic: "{bad}/{topic}"
            """)
        file_dict = {'zuul.d/test.yaml': in_repo_conf}
        A = self.fake_gerrit.addFakeChange('common-config', 'master', 'A',
                                           files=file_dict)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertIn("topic component 'bad' is invalid", A.messages[0],
                      "A should report a syntax error")

    @okay_tracebacks('Connection refused')
    def test_topic_replace_invalid_chars(self):
        # Test that special characters in the topic are replaced
        self.create_branch('org/project', 'weird#+')
        self.fake_gerrit.addEvent(
            self.fake_gerrit.getFakeBranchCreatedEvent(
                'org/project', 'weird#+'))
        self.waitUntilSettled()

        A = self.fake_gerrit.addFakeChange('org/project', 'weird#+', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        success_event = self.mqtt_messages.pop()
        self.assertEqual(success_event['topic'],
                         'tenant-one/zuul_buildset/check/org/project/weird__')


class TestElasticsearchConnection(AnsibleZuulTestCase):
    config_file = 'zuul-elastic-driver.conf'
    tenant_config_file = 'config/elasticsearch-driver/main.yaml'

    # These tests are storing the reported index on the fake
    # elasticsearch backend which is a different instance for each
    # scheduler. Thus, depending on which scheduler reports the
    # item, the assertions in these test might pass or fail.
    scheduler_count = 1

    def _getSecrets(self, job, pbtype):
        secrets = []
        build = self.getJobFromHistory(job)
        for pb in getattr(build.jobdir, pbtype):
            if pb.secrets_content:
                secrets.append(
                    yamlutil.ansible_unsafe_load(pb.secrets_content))
            else:
                secrets.append({})
        return secrets

    def test_elastic_reporter(self):
        "Test the Elasticsearch reporter"
        # Add a success result
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        indexed_docs = self.scheds.first.connections.connections[
            'elasticsearch'].source_it
        index = self.scheds.first.connections.connections[
            'elasticsearch'].index

        self.assertEqual(len(indexed_docs), 2)
        self.assertEqual(index, ('zuul-index.tenant-one-%s' %
                                 time.strftime("%Y.%m.%d")))
        buildset_doc = [doc for doc in indexed_docs if
                        doc['build_type'] == 'buildset'][0]
        self.assertEqual(buildset_doc['tenant'], 'tenant-one')
        self.assertEqual(buildset_doc['pipeline'], 'check')
        self.assertEqual(buildset_doc['result'], 'SUCCESS')
        build_doc = [doc for doc in indexed_docs if
                     doc['build_type'] == 'build'][0]
        self.assertEqual(build_doc['buildset_uuid'], buildset_doc['uuid'])
        self.assertEqual(build_doc['result'], 'SUCCESS')
        self.assertEqual(build_doc['job_name'], 'test')
        self.assertEqual(build_doc['tenant'], 'tenant-one')
        self.assertEqual(build_doc['pipeline'], 'check')

        self.assertIn('job_vars', build_doc)
        self.assertDictEqual(
            build_doc['job_vars'], {'bar': 'foo', 'bar2': 'foo2'})

        self.assertIn('job_returned_vars', build_doc)
        self.assertDictEqual(
            build_doc['job_returned_vars'], {'foo': 'bar'})

        self.assertEqual(self.history[0].uuid, build_doc['uuid'])
        self.assertIn('duration', build_doc)
        self.assertTrue(type(build_doc['duration']) is int)

        doc_gen = self.scheds.first.connections.connections[
            'elasticsearch'].gen(indexed_docs, index)
        self.assertIsInstance(doc_gen, types.GeneratorType)
        self.assertTrue('@timestamp' in list(doc_gen)[0]['_source'])

    def test_elasticsearch_secret_leak(self):
        expected_secret = [{
            'test_secret': {
                'username': 'test-username',
                'password': 'test-password'
            }
        }]

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        indexed_docs = self.scheds.first.connections.connections[
            'elasticsearch'].source_it

        build_doc = [doc for doc in indexed_docs if
                     doc['build_type'] == 'build'][0]

        # Ensure that job include secret
        self.assertEqual(
            self._getSecrets('test', 'playbooks'),
            expected_secret)

        # Check if there is a secret leak
        self.assertFalse('test_secret' in build_doc['job_vars'])


class TestConnectionsBranchCache(ZuulTestCase):
    config_file = "zuul-gerrit-github.conf"
    tenant_config_file = 'config/multi-driver/main.yaml'

    def test_branch_cache_fetch_error(self):
        # Test that a fetch error stores the right value in the branch cache
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        connection = self.scheds.first.connections.connections['github']
        source = connection.source
        project = source.getProject('org/project1')

        # Patch the fetch method so that it fails
        orig = connection._fetchProjectBranches

        def fail(*args, **kw):
            raise Exception("Unable to fetch branches")
        self.patch(connection, '_fetchProjectBranches', fail)

        # Clear the branch cache so we start with nothing
        connection.clearBranchCache()

        # Verify that we raise an error when we try to get branches
        # for a missing project
        self.assertRaises(
            Exception,
            lambda: connection.getProjectBranches(project, tenant))
        # This should happen again (ie, we should retry since we don't
        # have an entry)
        self.assertRaises(
            Exception,
            lambda: connection.getProjectBranches(project, tenant))

        # Restore the normal fetch method and verify that the cache
        # works as expected
        self.patch(connection, '_fetchProjectBranches', orig)
        branches = connection.getProjectBranches(project, tenant)
        self.assertEqual(['master'], branches)

        # Ensure that the empty list of branches is valid and is not
        # seen as an error
        self.init_repo("org/newproject")
        newproject = source.getProject('org/newproject')
        connection.addProject(newproject)
        tpc = zuul.model.TenantProjectConfig(newproject)
        tpc.exclude_unprotected_branches = True
        tenant.addTPC(tpc)
        branches = connection.getProjectBranches(newproject, tenant)
        self.assertEqual([], branches)
